#!/usr/bin/env python

"""
This module contains helper functions used internally in other fxmanager sub-packages. it also includes public functions to be used by the user to setup fxmanager project structure and pre-process the data.

Public Functions:
    - get_portfolios(): gets the portfolios to be used in historical backtesting of strategies.
    - setup()         : creates the required folders and files for any new project according to application type.
    - preprocess()    : preprocesses the raw data files to prepare it for analysis and simumlations.
"""

import sys
import json
from os import getcwd, mkdir, listdir
from os.path import join, isfile
from time import sleep
import operator as op
from functools import reduce
from shutil import copyfile
import inspect
from textwrap import dedent
import pandas as pd
import numpy as np
from fxmanager.strategies.template import strategy_template
import fxmanager._metadata as md
from __main__ import __dict__

__dict__.update(md.__dict__)

def get_avg_rets(day, data_dir=None, time_frames=[], currency_pairs=[]):
    """
    gets the average best & worst returns of the passed currency pairs & timeframes from the preprocessed data in the 'data_dir\\time_frames' directory.

    Args:
        - day           : integer indicating day number - starts from 0 and ends with the total number of work days in the dataset.
        - data_dir      : string with the directory or full path to the directory in which application data is kept. If the setup() function is used to create the recommended project structure, the default None value should be used.
        - time_frames   : a list of strings of the timeframes from which the data is read.
        - currency_pairs: a list of strings of the currency pairs from which the data is read.
    
    Returns:
        - avg_best_rets: pandas dataframe of average best return for each asset, columns are time_frames and index is currency_pairs.
        - avg_wrst_rets: pandas dataframe of average worst return for each asset, columns are time_frames and index is currency_pairs.
    
    Usage:
        - used in 'fxmanager.simulation.live.run()' function if the construct_portfolio is set to True.
        - used in 'fxmanager.optimization.w_optimization_test' and 'fxmanager.optimization.eqw_optimization_test' modules.
    """

    if data_dir is None:
        data_dir = join(getcwd(), 'data')
    avg_best_rets = pd.DataFrame()
    avg_wrst_rets = pd.DataFrame()
    for i in time_frames:
        tf_rets = pd.DataFrame()
        for cp in currency_pairs:
            df = pd.read_csv(join(data_dir, 'time_frames', str(day), i , cp+'_'+i+'.csv'), parse_dates=True) 
            # Calculate average returns
            comp_best_rets = (df['best_rets']).mean()
            comp_wrst_rets = (df['wrst_rets']).mean()
            
            tf_rets[cp] = pd.Series({'comp_best_rets':comp_best_rets,
                                     'comp_wrst_rets':comp_wrst_rets
                                    })
        avg_best_rets[i] = tf_rets.loc['comp_best_rets']
        avg_wrst_rets[i] = tf_rets.loc['comp_wrst_rets']
        
    return avg_best_rets, avg_wrst_rets

def get_portfolios(data_dir=None, num_days=1, single_portfolio={}, portfolios=[], optimized_portfolios=False, use_single_porfolio=False):
    """
    gets the portfolios to be used in historical backtesting of strategies.

    Args:
        - data_dir        : string with the directory or full path to the directory in which application data is kept. If the setup() function is used to create the recommended project structure, the default None value should be used.
        - num_days        : total number of work days in the dataset.
        - single_portfolio: dictionary of the portfolio to be used for backtesting with keys (currency_pairs, time_frames, weights) used if 'optimized_portfolios' is False and 'use_single_porfolio' is True.
        - portfolios      : list of dictionries like 'single_portfolio'. Lenght of the list = num_days-1. used if 'optimized_portfolios' is False and 'use_single_porfolio' is False.
        - optimized_portfolios: boolean flag, set to True if an optimization test is run and the portfolios to be used are saved in 'data_dir\\stats\\Optimizer_Test_portfolios.csv' file.
        - use_single_porfolio: boolean flag, used if 'optimized_portfolios' is False, set to true if you passed a single portfolio in the 'single_portfolio' argument, if False, 'portfolios' argument is used instead.
    
    Returns:
        - df: pandas dataframe with columns (currency_pairs, time_frames, weights) and range index of length = num_days
    
    Usage:
        - used by user before calling the function 'fxmanager.simulation.historic.run()'. 
    """

    if data_dir is None:
        data_dir = join(getcwd(), 'data')
    if optimized_portfolios:
        df = pd.read_csv(join(data_dir, 'stats', 'Optimizer_Test_portfolios.csv'), parse_dates=True)
        currency_pairs, time_frames, weights = [], [], []
        for idx in df.index:
            cps = [cp.strip('\'') for cp in df.loc[idx, 'currency_pairs'].lstrip('[').rstrip(']').split(',')]
            tfs = [tf.strip('\'') for tf in df.loc[idx, 'time_frames'].lstrip('[').rstrip(']').split(',')]
            w = [float(w) for w in df.loc[idx, 'weights'].lstrip('[').rstrip(']').split(',')]
            currency_pairs.append(cps)
            time_frames.append(tfs)
            weights.append(w)
        df = pd.DataFrame({'currency_pairs': currency_pairs,
                           'time_frames': time_frames,
                           'weights': weights
                          })
        df.index.name='days'
        
    else:
        if use_single_porfolio:
            df = pd.DataFrame([single_portfolio]*num_days)
            df.index.name='days'
        else:
            df = pd.DataFrame(portfolios)
            df.index.name='days'
    return df

def get_weights(data_dir=None, num_assets=1):
    """
    reads the pre-calculated weights from 'data_dir\\weights\\' directory to be used for weight optimization.

    Args:
        - data_dir   : string with the directory or full path to the directory in which application data is kept. If the setup() function is used to create the recommended project structure, the default None value should be used.
        - num_assets : number of assets in the portfolio being optimized. An asset is a currency pair in a specific timeframe.
    
    Returns:
        - weight_combinations_list: list of all possible weight combinations for the passed number of assets.
    
    Usage:
        - used in 'fxmanager.optimization.weight_optimizer' module.
    """

    if data_dir is None:
        data_dir = join(getcwd(), 'data')
    weight_combinations_list = []
    weight_file_name = str(num_assets)+'_asset_weights.txt'
    with open(join(data_dir,'weights',weight_file_name),'r') as f:
        lines = f.readlines()
        for line in lines:
            line = [float(w) for w in line[:-2].split(',')]
            while len(line) < num_assets:
                line.append(0)
            weight_combinations_list.append(line)
    
    return weight_combinations_list

def top_n_timeframes(best, wrst, objective='min_win_rate', n=1):
    """
    selects the best (n) of time frames according to the 'objective' argument.

    Args:
        - best     : pandas dataframe with the return of the 'get_avg_rets()' function defined above. note tha the index must be transformed first to range index.
        - wrst     : pandas dataframe with the return of the 'get_avg_rets()' function defined above. note tha the index must be transformed first to range index.
        - objective: string indecating the criteria with which the best time frames will be chosed, supported objectives are (min_win_rate, min_risk_reward)
        - n        : integer indicating the number of time frames to return
    
    Returns:
        - top_n: numpy array with the names of selected best time frames.
        - W_R  : pandas series with required win rates for every time frame.
        - R_R  : pandas series with required risk rewards for every time frame.
    
    Usage:
        - used in 'fxmanager.optimization.weight_optimizer' and 'fxmanager.optimization.eq_weight_optimizer' modules.
    """

    num_cps = best.shape[0]

    weights = [1/num_cps,] * num_cps
    weights_df = pd.DataFrame(weights,columns=['W'])
    W_S_best = best.multiply(weights_df['W'], axis=0).sum(axis=0)
    W_S_wrst = wrst.multiply(weights_df['W'], axis=0).sum(axis=0)
    W_R = (abs(W_S_wrst) / (W_S_best + abs(W_S_wrst)))
    R_R = abs(W_S_wrst) / W_S_best

    if objective == 'min_win_rate':
        top_n = W_R.nsmallest(n).index.values
        return top_n, W_R, R_R
    elif objective == 'min_risk_reward':
        top_n = R_R.nsmallest(n).index.values
        return top_n, W_R, R_R

def findCombinationsUtil(arr, index, num, reducedNum):
    """
    helper function for findCombinations() function.
    """

    # Base condition
    if (reducedNum < 0):
        return

    if (reducedNum == 0):
        for i in range(index):
            print(arr[i], end = ",")
        print("")
        return
 
    prev = 1 if(index == 0) else arr[index - 1]
 
    for k in range(prev, num + 1):
        arr[index] = k
        findCombinationsUtil(arr, index + 1, num,
                                 reducedNum - k)
 
def findCombinations(n):
    """
    finds all possible combinations that sum up to n.

    Args:
        - n: integer idicating the number that returned cambinations sum up to.
    
    Returns:
        - every combination is printed in a new line to standard output. to save the results to a file, change sys.stdout to the path of the file.
    
    Usage:
        - used in 'fxmanager.basic.util.setup()' function to save required weight combinations to 'data_dir\\weigts' directory.
    """

    arr = [0] * n
    findCombinationsUtil(arr, 0, n, n)

def ncr(n, r):
    """
    calculates the number of combinations nCrز
    
    Args:
        - n: integer indicating the total number of items to be combined.
        - r: integer indicating the length of every combination.
    
    Returns:
        - res: integer with the result.
    """

    r = min(r, n-r)
    numer = reduce(op.mul, range(n, n-r, -1), 1)
    denom = reduce(op.mul, range(1, r+1), 1)
    res = numer // denom
    return res

def expected_return(day=0, data_dir=None, time_frames=[], currency_pairs=[], weights=[], leverage=0.01, win_rate=0.5, risk_reward=1, n_scenarios=100):
    """
    calculates the average expected daily return from a portfolio given a win rate and a risk-reward ratio.

    Args:
        - day           : integer with range from 0 to (number of working days in the dataset-1).
        - data_dir      : string with the directory or full path to the directory in which application data is kept. If the setup() function is used to create the recommended project structure, the default None value should be used.
        - time_frames   : list of portfolio timeframes.
        - currency_pairs: list of portfolio currency pairs.
        - weights       : list of portfolio weights.
        - leverage      : float indicating the leverage. if no leverage is used set to 1.
        - win_rate      : float indicating the percentage of winning positions to the total number of positions.
        - risk_reward   : float indicating ratio of reward to risk.
        - n_scenarios   : integer indicating the number of scenarios across which the expected return is averaged.
    
    Returns:
        - avg_expected_return: float indicating the average expected return.
        - win_probability    : float indicating the ratio of number of scenarios in which the expected return is positive to the total number
                               of scenarios.
    
    Usage:
        - used in 'fxmanage.optimization.eqw_optimization_test.run_test()' function to calculate the expected returns using the optimized 
          portfolios.
    """

    if data_dir is None:
        data_dir = join(getcwd(), 'data')
    if win_rate > 1:
        win_rate = 1
    loss_rate = 1 - win_rate
    num_assets = len(time_frames) # or len(currency_pairs), both are the same

    er = 0
    win_probability = 0
    for i in range(n_scenarios):
        scenario_rets = 0
        for tf, cp, weight in zip(time_frames, currency_pairs, weights):
            file_name = cp + '_' + tf + '.csv'
            df = pd.read_csv(join(data_dir, 'time_frames', str(day), tf, file_name))
            df.loc[:, ('best_rets', 'wrst_rets')] /= leverage
            num_wins = int(win_rate * df.shape[0])
            num_losses = df.shape[0] - num_wins
            if num_wins > df.shape[0]:
                num_wins = df.shape[0]
                num_losses = 0
            if risk_reward > 1: ## Risk is more than reward >> take 100% from loss samples and 1/risk_reward from win samples
                sampled_win_rets = df['best_rets'].sample(num_wins)*weight*(1/risk_reward)
                sampled_loss_rets = df['wrst_rets'].sample(num_losses)*weight
            elif risk_reward < 1:
                sampled_win_rets = df['best_rets'].sample(num_wins)*weight
                sampled_loss_rets = df['wrst_rets'].sample(num_losses)*weight*risk_reward
            else:
                sampled_win_rets = df['best_rets'].sample(num_wins)*weight
                sampled_loss_rets = df['wrst_rets'].sample(num_losses)*weight

            rets = pd.concat([sampled_win_rets, sampled_loss_rets], ignore_index=True)
            scenario_rets += (1+rets).prod()-1

        er += scenario_rets
        if scenario_rets > 0:
            win_probability += (1/n_scenarios)
    avg_expected_return = er/n_scenarios
    return avg_expected_return, win_probability

def setup(app_type='', raw_data_dir='', live_raw_data_dir='', raw_data_type='', file_name_format='', num_days=1, num_cps=1, num_tfs=1):
    """
    creates the required folders and files for any new project according to application type.

    Args:
        - app_type : string indicating the application type. supported app types are:
            - back_tester          : if you want to test trading strategies using histoical data on FX-Manager virtual account.
            - portfolio_optimizer  : if you want to asses different portfolio optimization methods and objectives using histoical data.
            - live_simulator       : if you want to test trading strategies in real time using MT4 Live data on FX-Manager virtual account.
            - all_in_one           : if you want to use all the previous features.
        
        - raw_data_dir     : string containig the path to directory in which the raw data files are kept.
        - live_raw_data_dir: string containig the path to directory in which the raw data files from the previous day are kept. used to construct a portfolio using built-in fxmanager optimization for live simulation.
        
        - raw_data_type : string with the type of raw data used, supported raw data types are: 
            - daily_bid_ask  : Bid/Ask ticks in daily seperate files (a file for each currency_pair in each day)
            - daily_candles  : 1-min candles in daily seperate files (a file for each currency_pair in each day)
            - period_candles : 1-min candles over a period (a file for each currency_pair for all the days)
        
        - file_name_format : string of white spaces with the same length as raw data file names. Givin that each file name has information about date and currency pair symbol, the white spaces in the same location as these information are replaced as follows:
            - day, month and year in the date are replaced with (dd), (mm), (yyyy) respectively.
            - currency pair symbol is replaced with (cccccc). For example, if the file name is 'AUDUSD_Ticks_02.08.2021-02.08.2021.csv' then the file_name_format is: 'cccccc       dd mm yyyy               '
                                         
        - num_days         : total number of work days in the dataset.
        - num_cps          : integer indicating the number of currency pairs in the dataset
        - num_tfs          : integer indicating the number of timeframes in the dataset.
    
    """

    print('\n>> SYSTEM MESSAGE >> STAGE 1: Creating Necessary Folders And Files ..\n')

    cwd = getcwd()
    raw_data_formats = {
        "daily_bid_ask" : {
            "date_col"    : "Gmt time",
            "date_format" : "%d.%m.%Y %H:%M:%S.%f",
            "ask_col"     : "Ask",
            "bid_col"     : "Bid"
        },

        "daily_candles" : {
            "date_col"      : "time",
            "date_format"   : "%Y.%m.%d %H:%M",
            "open_col"      : "open",
            "high_col"      : "high",
            "low_col"       : "low",
            "close_col"     : "close",
            "tick_vol_col"  : "tick_volume"
        },

        "period_candles" : {
            "date_col"       : "time",
            "date_format"    : "%Y.%m.%d %H:%M",
            "open_col"       : "open",
            "high_col"       : "high",
            "low_col"        : "low",
            "close_col"      : "close",
            "tick_vol_col"   : "tick_volume"
        }  
    }
    json_object = json.dumps(raw_data_formats, indent = 4)
    with open("raw_data_formats.json", "w") as outfile:
        outfile.write(json_object)
    
    # create strategy.py file
    lines = inspect.getsource(strategy_template.template)
    with open('strategy.py', 'w') as f:
        f.write(dedent(lines))
    
    # add keyword arguments to strategy.py file
    lines = inspect.getsource(strategy_template.kwargs_template)
    with open('strategy.py', 'a') as f:
        f.write(dedent(lines))
    
    # add portfilios to strategy.py file
    lines = inspect.getsource(strategy_template.get_portfolios_template)
    with open('strategy.py', 'a') as f:
        f.write(dedent(lines))
    
    mkdir(join(cwd,'data'))
    mkdir(join(cwd,'data', 'logs'))
    mkdir(join(cwd,'data', 'visualizations'))
    mkdir(join(cwd,'data', 'stats'))
    mkdir(join(cwd,'data', 'raw_data'))
    mkdir(join(cwd,'data', 'raw_data_archive'))
    mkdir(join(cwd,'data', 'time_frames'))
    mkdir(join(cwd,'data', 'weights'))
    
    for i in range(1, max(num_cps, num_tfs)+1):
        sys.stdout = open(join(cwd,'data', 'weights', str(i)+'_asset_weights.txt'), 'w')
        findCombinations(i)
        sys.stdout.close()
    sys.stdout = sys.__stdout__
    
    if app_type == 'back_tester' or app_type == 'portfolio_optimizer' or app_type == 'all_in_one':
        for i in range(num_days):
            mkdir(join(cwd,'data', 'time_frames', str(i)))
            for j in range(1,num_tfs+1):
                mkdir(join(cwd,'data', 'time_frames', str(i), str(j)+'min'))
        if app_type == 'back_tester':
            mkdir(join(cwd,'data', 'stats', 'historical_simulation_orders'))
        if app_type == 'all_in_one':
            mkdir(join(cwd,'data', 'stats', 'historical_simulation_orders'))
            mkdir(join(cwd,'data', 'live_data'))
            mkdir(join(cwd,'data', 'live_data', 'raw_data'))
            mkdir(join(cwd,'data', 'live_data', 'time_frames'))
            for i in range(1,num_tfs+1):
                mkdir(join(cwd,'data', 'live_data', 'time_frames', str(i)+'min'))
            
            raw_file_names = [f for f in listdir(live_raw_data_dir)]
            for file_name in raw_file_names:
                currency_pair = ''.join([file_name[n] for n,m in enumerate(file_name_format) if m =='c'])
                new_file_name = currency_pair + '.csv'
                copyfile(join(live_raw_data_dir,file_name), join(cwd,'data', 'live_data', 'raw_data', new_file_name))

        print('>> SYSTEM MESSAGE >> All Necessary Folders And Files Are Created Successfully! ..\n')
        print('>> SYSTEM MESSAGE >> STAGE 2: Coppying Raw Data Files ..\n')
        
        raw_file_names = [f for f in listdir(raw_data_dir)]
        if raw_data_type == 'daily_bid_ask' or raw_data_type == 'daily_candles':
            dates = []
            for file_name in raw_file_names:
                day = ''.join([file_name[i] for i,m in enumerate(file_name_format) if m =='d'])
                month = ''.join([file_name[i] for i,m in enumerate(file_name_format) if m =='m'])
                year = ''.join([file_name[i] for i,m in enumerate(file_name_format) if m =='y'])
                dates.append(month+'.'+day+'.'+year)
            sorted_file_names = pd.DataFrame(raw_file_names, index=pd.to_datetime(dates), columns=[0]).sort_index()
            unique_dates = set(dates)
            unique_dates_sorted = pd.DataFrame([0]*num_days, index=pd.to_datetime(list(unique_dates))).sort_index().index
                        
            for i, idx in enumerate(unique_dates_sorted):
                print(f'>> PROCESS MESSAGE >> Coppying Data Of Day {i} ..')
                mkdir(join(cwd,'data', 'raw_data', str(i)))
                for file_name in sorted_file_names.loc[idx,0]:
                    if num_cps == 1:
                        file_name = sorted_file_names.loc[idx,0]
                    # move files
                    currency_pair = ''.join([file_name[n] for n,m in enumerate(file_name_format) if m =='c'])
                    day = ''.join([file_name[i] for i,m in enumerate(file_name_format) if m =='d'])
                    month = ''.join([file_name[i] for i,m in enumerate(file_name_format) if m =='m'])
                    year = ''.join([file_name[i] for i,m in enumerate(file_name_format) if m =='y'])
                    new_file_name = currency_pair + '_' + str(pd.to_datetime([month+'.'+day+'.'+year])[0].date()) + '.csv'
                    copyfile(join(raw_data_dir,file_name), join(cwd,'data', 'raw_data', str(i), new_file_name))
                    if num_cps == 1:
                        break 
        
        elif raw_data_type == 'period_candles':
            for file_name in raw_file_names:
                currency_pair = ''.join([file_name[n] for n,m in enumerate(file_name_format) if m =='c'])
                new_file_name = currency_pair + '.csv'
                copyfile(join(raw_data_dir,file_name), join(cwd,'data', 'raw_data', new_file_name))
        else:
            pass
        
        print('\n>> SYSTEM MESSAGE >> All Raw Data Files Are Coppied Successfully! ..\n')
    
    elif app_type == 'live_simulator':
        for i in range(1,num_tfs+1):
            mkdir(join(cwd,'data', 'time_frames', str(i)+'min'))
        
        print('>> SYSTEM MESSAGE >> All Necessary Folders Are Created Successfully! ..\n')
        print('>> SYSTEM MESSAGE >> STAGE 2: Coppying Raw Data Files ..\n')

        raw_file_names = [f for f in listdir(live_raw_data_dir)]
        for file_name in raw_file_names:
            currency_pair = ''.join([file_name[n] for n,m in enumerate(file_name_format) if m =='c'])
            new_file_name = currency_pair + '.csv'
            copyfile(join(live_raw_data_dir,file_name), join(cwd,'data', 'raw_data', new_file_name))

        print('>> SYSTEM MESSAGE >> All Raw Data Files Are Coppied Successfully! ..\n')
  
    else:
        pass
    
    print('>> SYSTEM MESSAGE >> Finishing ..\n')

def process_daily_bid_ask(df, tf):
    """
    helper function for 'fxmanager.basic.util.preprocess()' function.
    
    used to group the given Bid/Ask data by the given timeframe and calculate open and close prices for the new candle sticks in the new timeframe.
    """
    # initialize placeholders for final data
    ask_open = []
    bid_open = []
    ask_close = []
    bid_close = []
    ask_high = []
    bid_high = []
    ask_low	= []
    bid_low = []
    ask_sample1 = []
    bid_sample1 = []
    ask_sample2 = []
    bid_sample2 = []
    best_rets = []
    wrst_rets = []
    tick_volume = []

    # get a DataFrame that is grouped by the specified timeframe
    grouped = pd.DataFrame(df.groupby(pd.Grouper(freq=tf)), columns=['time', 'sub_dataframes'])

    for dataframe in grouped['sub_dataframes']:
        tick_volume.append(dataframe.shape[0])
        if dataframe.shape[0] != 0:
            ask_open.append(dataframe['ask'][0])
            bid_open.append(dataframe['bid'][0])
            ask_close.append(dataframe['ask'][-1])
            bid_close.append(dataframe['bid'][-1])
            ask_high.append(dataframe['ask'].max())
            bid_high.append(dataframe['bid'].max())
            ask_low.append(dataframe['ask'].min())
            bid_low.append(dataframe['bid'].min())
            
            delta_t = (dataframe.index[-1]-dataframe.index[0])/2
            open_plus_delta = delta_t + dataframe.index[0]
            idx_diff = dataframe.index - open_plus_delta
            pos_idx_diff = [a for a in idx_diff if a.days==0]
            idx = min(pos_idx_diff) + open_plus_delta

            ask_sample1.append(dataframe.loc[dataframe.index[0]:idx]['ask'].sample().values[0])
            bid_sample1.append(dataframe.loc[dataframe.index[0]:idx]['bid'].sample().values[0])
            ask_sample2.append(dataframe.loc[idx:dataframe.index[-1]]['ask'].sample().values[0])
            bid_sample2.append(dataframe.loc[idx:dataframe.index[-1]]['bid'].sample().values[0])
            
            if bid_high[-1] > ask_low[-1]:
                ## if minimum ask price comes first >> buy >> devide by ask_low
                if max(dataframe['bid'].idxmax(), dataframe['ask'].idxmin()) == dataframe['bid'].idxmax():
                    best_rets.append((bid_high[-1]-ask_low[-1]) / ask_low[-1])
                    wrst_rets.append((bid_low[-1]-ask_high[-1]) / bid_low[-1])
                ## if maximum bid price comes first >> sell >> devide by bid_high
                else:
                    best_rets.append((bid_high[-1]-ask_low[-1]) / bid_high[-1])
                    wrst_rets.append((bid_low[-1]-ask_high[-1]) / ask_high[-1])

            else:
                best_rets.append(0)
                if max(dataframe['bid'].idxmin(), dataframe['ask'].idxmax()) == dataframe['bid'].idxmin():
                    wrst_rets.append((bid_low[-1]-ask_high[-1]) / bid_low[-1])
                else:
                    wrst_rets.append((bid_low[-1]-ask_high[-1]) / ask_high[-1])
            
        else:
            ask_open.append(ask_close[-1])
            bid_open.append(bid_close[-1])
            ask_close.append(ask_close[-1])
            bid_close.append(bid_close[-1])
            ask_high.append(ask_close[-1])
            bid_high.append(bid_close[-1])
            ask_low.append(ask_close[-1])
            bid_low.append(bid_close[-1])
            ask_sample1.append(ask_close[-1])
            bid_sample1.append(bid_close[-1])
            ask_sample2.append(ask_close[-1])
            bid_sample2.append(bid_close[-1])
            best_rets.append(0)
            wrst_rets.append((bid_low[-1]-ask_high[-1]) / bid_low[-1])

    final_df = pd.DataFrame({'best_rets':best_rets,
                             'wrst_rets':wrst_rets,
                             'ask_open':ask_open,
                             'bid_open':bid_open,
                             'ask_close':ask_close,
                             'bid_close':bid_close,
                             'ask_sample1':ask_sample1,
                             'bid_sample1':bid_sample1,
                             'ask_sample2':ask_sample2,
                             'bid_sample2':bid_sample2,
                             'tick_volume':tick_volume
                            }, index=pd.to_datetime(grouped['time']))
    return final_df

def process_ohlc(df, tf):
    """
    helper function for 'fxmanager.basic.util.preprocess()' function.

    used to group the given OHLCV data by the given timeframe and calculate open and close prices for the new candle sticks in the new timeframe.
    """

    # initialize placeholders for final data
    ask_open = []
    bid_open = []
    ask_close = []
    bid_close = []
    ask_high = []
    bid_high = []
    ask_low	= []
    bid_low = []
    ask_sample1 = []
    bid_sample1 = []
    ask_sample2 = []
    bid_sample2 = []
    best_rets = []
    wrst_rets = []
    tick_volume = []

    # Calculate spread using Corwin-Schultz model
    l = df['low']
    h = df['high']
    gamma = (np.log(h.rolling(2).max()/l.rolling(2).min()))**2
    beta = (np.log(h/l))**2 + (np.log(h.shift(1)/l.shift(1)))**2
    alpha = ((np.sqrt(2*beta) - np.sqrt(beta)) / (3-(2*np.sqrt(2)))) - np.sqrt(gamma / (3-(2*np.sqrt(2))))
    s = 2*(np.exp(alpha)-1) / (1+np.exp(alpha))
    s = abs(s).round(5)
    s[0] = s[1]

    # Add bid/ask column to df 
    df = pd.DataFrame({'bid_open': df['open'].values,
                       'ask_open': (df['open'].to_numpy()+s).round(5),
                       'bid_close': df['close'].values,
                       'ask_close': (df['close'].to_numpy()+s).round(5),
                       'bid_high': df['high'].values,
                       'ask_high': (df['high'].to_numpy()+s).round(5),
                       'bid_low': df['low'].values,
                       'ask_low': (df['low'].to_numpy()+s).round(5),
                       'tick_volume':df['tick_volume'].values}, index=df.index)
    # get a DataFrame that is grouped by the specified timeframe
    grouped = pd.DataFrame(df.groupby(pd.Grouper(freq=tf)), columns=['time', 'sub_dataframes'])

    for dataframe in grouped['sub_dataframes']:
        tick_volume.append(dataframe['tick_volume'].sum())
        ask_open.append(dataframe['ask_open'][0])
        bid_open.append(dataframe['bid_open'][0])
        ask_close.append(dataframe['ask_close'][-1])
        bid_close.append(dataframe['bid_close'][-1])
        ask_high.append(dataframe['ask_high'].max())
        bid_high.append(dataframe['bid_high'].max())
        ask_low.append((dataframe['ask_low'].min()))
        bid_low.append(dataframe['ask_low'].min())
        
        delta_t = (dataframe.index[-1]-dataframe.index[0])/2
        open_plus_delta = delta_t + dataframe.index[0]
        idx_diff = dataframe.index - open_plus_delta
        pos_idx_diff = [a for a in idx_diff if a.days==0]
        idx = min(pos_idx_diff) + open_plus_delta

        ask_sample1.append(dataframe.loc[:idx]['ask_open'].sample().values[0])
        bid_sample1.append(dataframe.loc[:idx]['bid_open'].sample().values[0])
        ask_sample2.append(dataframe.loc[idx:]['ask_close'].sample().values[0])
        bid_sample2.append(dataframe.loc[idx:]['bid_close'].sample().values[0])
        
        if bid_high[-1] > ask_low[-1]:
            ## if minimum ask price comes first >> buy >> devide by ask_low
            if max(dataframe['bid_high'].idxmax(), dataframe['ask_low'].idxmin()) == dataframe['bid_high'].idxmax():
                best_rets.append((bid_high[-1]-ask_low[-1]) / ask_low[-1])
                wrst_rets.append((bid_low[-1]-ask_high[-1]) / bid_low[-1])
            ## if maximum bid price comes first >> sell >> devide by bid_high
            else:
                best_rets.append((bid_high[-1]-ask_low[-1]) / bid_high[-1])
                wrst_rets.append((bid_low[-1]-ask_high[-1]) / ask_high[-1])
        else:
            best_rets.append(0)
            # print(dataframe, end='\n\n')
            if max(dataframe['bid_low'].idxmin(), dataframe['ask_high'].idxmax()) == dataframe['bid_low'].idxmin():
                wrst_rets.append((bid_low[-1]-ask_high[-1]) / bid_low[-1])
            else:
                wrst_rets.append((bid_low[-1]-ask_high[-1]) / ask_high[-1])

    final_df = pd.DataFrame({'best_rets':best_rets,
                             'wrst_rets':wrst_rets,
                             'ask_open':ask_open,
                             'bid_open':bid_open,
                             'ask_close':ask_close,
                             'bid_close':bid_close,
                             'ask_sample1':ask_sample1,
                             'bid_sample1':bid_sample1,
                             'ask_sample2':ask_sample2,
                             'bid_sample2':bid_sample2,
                             'tick_volume':tick_volume
                            }, index=pd.to_datetime(grouped['time']))
    return final_df

def preprocess(data_dir=None, app_type='', raw_data_type='', raw_data_format={}, save_logs=True):
    """
    preprocesses the raw data files to prepare it for analysis and simumlations.

    Args:
        - data_dir : string with the directory or full path to the directory in which application data is kept. If the setup() function is used to create the recommended project structure, the default None value should be used.

        - app_type : string indicating the application type. supported app types are
            - back_tester          : if you want to test trading strategies using histoical data on FX-Manager virtual account.
            - portfolio_optimizer  : if you want to asses different portfolio optimization methods and objectives using histoical data.
            - live_simulator       : if you want to test trading strategies in real time using MT4 Live data on FX-Manager virtual account.
            - all_in_one           : if you want to use all the previous features.
        
        - raw_data_type : string with the type of raw data used, supported raw data types are
            - daily_bid_ask  : Bid/Ask ticks in daily seperate files (a file for each currency_pair in each day)
            - daily_candles  : 1-min candles in daily seperate files (a file for each currency_pair in each day)
            - period_candles : 1-min candles over a period (a file for each currency_pair for all the days)

        - raw_data_format : dictionary with information about row data files, the required keys vary according to raw_data_type argument, the following are exambles of how the dictionary should be formatted for each type
            - daily_bid_ask: {'date_col':'Gmt Time', 'date_format':'%d.%m.%Y %H:%M:%S.%f', 'ask_col':'Ask', 'bid_col':'Bid'}
            - daily_candles or period_candles: {'date_col':'time', 'date_format':'%Y.%m.%d %H:%M', 'open_col':'open', 'high_col':'high', 'low_col':'low', 'close_col':'close', 'tick_vol_col':'tick_volume'}
        
        - save_logs : boolean flag, if True, the program logs are saved to 'data_dir\\logs\\preprocessing_logs.txt' file.
    Returns:
        - None
    """

    if data_dir is None:
        data_dir = join(getcwd(), 'data')
    if save_logs:
        sys.stdout = open(join(data_dir, 'logs','preprocessing_logs.txt'), 'w')

    #############################################################################################################################
    ##                                                  Daily Bid/Ask Data                                                     ##
    #############################################################################################################################

    if raw_data_type == 'daily_bid_ask':
        # Extract formatting information 
        date_col = raw_data_format['date_col']
        date_format = raw_data_format['date_format']
        ask_col = raw_data_format['ask_col']
        bid_col = raw_data_format['bid_col']

        #############################################################################################################################
        
        if app_type == 'back_tester' or app_type == 'portfolio_optimizer' or app_type == 'all_in_one':
            # Find num_days and time_frames
            num_days = len(listdir(join(data_dir, 'time_frames')))
            tf_list = listdir(join(data_dir, 'time_frames', '0'))
            time_frames = []
            for i in range(1, len(tf_list)+1):
                for tf in tf_list:
                    if tf[1].isalpha():
                        x = int(tf[0])
                    else:
                        x = int(tf[:2])
                    if x == i:
                        time_frames.append(tf)
            for day in range(num_days):
                print('-----------------------------------------------------------------------------------------------------------')
                print('                                             >> Processing Day {} <<'.format(day))
                print('-----------------------------------------------------------------------------------------------------------')

                for i in time_frames:
                    print('\n>> PROCESS MESSAGE >> Processing {} Timeframe ..'.format(i))
                    file_names = [f for f in listdir(join(data_dir, 'raw_data', str(day)))]
                    for file_name in file_names:
                        print('>> PROCESS MESSAGE >>     Processing {} Currency pair ..'.format(file_name[:6]))
                        df = pd.read_csv(join(data_dir, 'raw_data', str(day), file_name))
                        df.index = pd.to_datetime(df[date_col], format=date_format)
                        df = df[[ask_col, bid_col]]
                        df.columns = ['ask', 'bid']
                        df = process_daily_bid_ask(df=df, tf=i)
                        df.to_csv(join(data_dir, 'time_frames', str(day), i, file_name[:6]+'_'+i+'.csv'))
            print('\n>> SYSTEM MESSAGE >> all Days Are Processed Successfully!\n')            
            
            if app_type == 'all_in_one':
                print('>> SYSTEM MESSAGE >> Processing Live App Data!')
                for i in time_frames:
                    print('\n>> PROCESS MESSAGE >> Processing {} Timeframe ..'.format(i))
                    file_names = [f for f in listdir(join(data_dir, 'live_data', 'raw_data'))]
                    for file_name in file_names:
                        print('>> PROCESS MESSAGE >>     Processing {} Currency pair ..'.format(file_name[:6]))
                        df = pd.read_csv(join(data_dir, 'live_data', 'raw_data', file_name))
                        df.index = pd.to_datetime(df[date_col], format=date_format)
                        df = df[[ask_col, bid_col]]
                        df.columns = ['ask', 'bid']
                        df = process_daily_bid_ask(df=df, tf=i)
                        df.to_csv(join(data_dir, 'live_data', 'time_frames', i, file_name[:6]+'_'+i+'.csv'))

        #############################################################################################################################    
        
        elif app_type == 'live_simulator':
            tf_list = listdir(join(data_dir, 'time_frames'))
            time_frames = []
            for i in range(1, len(tf_list)+1):
                for tf in tf_list:
                    if tf[1].isalpha():
                        x = int(tf[0])
                    else:
                        x = int(tf[:2])
                    if x == i:
                        time_frames.append(tf)
            print('>> SYSTEM MESSAGE >> Processing Live App Data!')
            for i in time_frames:
                print('\n>> PROCESS MESSAGE >> Processing {} Timeframe ..'.format(i))
                file_names = [f for f in listdir(join(data_dir, 'raw_data'))]
                for file_name in file_names:
                    print('>> PROCESS MESSAGE >>     Processing {} Currency pair ..'.format(file_name[:6]))
                    df = pd.read_csv(join(data_dir, 'raw_data', file_name))
                    df.index = pd.to_datetime(df[date_col], format=date_format)
                    df = df[[ask_col, bid_col]]
                    df.columns = ['ask', 'bid']
                    df = process_daily_bid_ask(df=df, tf=i)
                    df.to_csv(join(data_dir, 'time_frames', i, file_name[:6]+'_'+i+'.csv'))
                
    #############################################################################################################################
    ##                                                  OHLCV 1min-Candles Data                                                ##
    #############################################################################################################################
    
    elif raw_data_type == 'daily_candles' or raw_data_type == 'period_candles':
        # Extract formatting information 
        date_col = raw_data_format['date_col']
        date_format = raw_data_format['date_format']
        open_col = raw_data_format['open_col']
        close_col = raw_data_format['close_col']
        low_col = raw_data_format['low_col']
        high_col = raw_data_format['high_col']
        tick_vol_col = raw_data_format['tick_vol_col']

        #############################################################################################################################

        if app_type == 'back_tester' or app_type == 'portfolio_optimizer' or app_type == 'all_in_one':
            # Find num_days and time_frames
            num_days = len(listdir(join(data_dir, 'time_frames')))
            tf_list = listdir(join(data_dir, 'time_frames', '0'))
            time_frames = []
            for i in range(1, len(tf_list)+1):
                for tf in tf_list:
                    if tf[1].isalpha():
                        x = int(tf[0])
                    else:
                        x = int(tf[:2])
                    if x == i:
                        time_frames.append(tf)
            for day in range(num_days):
                print('-----------------------------------------------------------------------------------------------------------')
                print('                                             >> Processing Day {} <<'.format(day))
                print('-----------------------------------------------------------------------------------------------------------')

                for i in time_frames:
                    print('\n>> PROCESS MESSAGE >> Processing {} Timeframe ..'.format(i))
                    if raw_data_type == 'daily_candles':
                        file_names = [f for f in listdir(join(data_dir, 'raw_data', str(day)))]
                    else:
                        file_names = [f for f in listdir(join(data_dir, 'raw_data'))]
                    for file_name in file_names:
                        print('>> PROCESS MESSAGE >>     Processing {} Currency pair ..'.format(file_name[:6]))
                        if raw_data_type == 'daily_candles':
                            df = pd.read_csv(join(data_dir, 'raw_data', str(day), file_name))
                        else:
                            df = pd.read_csv(join(data_dir, 'raw_data', file_name))
                        df = pd.DataFrame({'open':df[open_col].values,
                                           'high':df[high_col].values,
                                           'low':df[low_col].values,
                                           'close':df[close_col].values,
                                           'tick_volume':df[tick_vol_col].values}, index=pd.to_datetime(df[date_col], format=date_format))
                        if raw_data_type == 'daily_candles':
                            idx = pd.date_range(df.index.date[0], freq='min', periods=1440)
                            df = df.reindex(idx, method='ffill')
                            df = df.replace(to_replace=np.nan, method='bfill')
                            df = process_ohlc(df=df, tf=i)
                        else:
                            idxs = pd.date_range(df.index.date[0], freq='B', periods=num_days)
                            idx = pd.date_range(idxs[0], freq='min', periods=1440)
                            for date in idxs[1:]:
                                idx = idx.union(pd.date_range(date, freq='min', periods=1440))
                            df = df.reindex(idx, method='ffill')
                            df = df.replace(to_replace=np.nan, method='bfill')
                            df = df.iloc[day*1440:(day+1)*1440]
                            df = process_ohlc(df=df, tf=i)
                        df.to_csv(join(data_dir, 'time_frames', str(day), i, file_name[:6]+'_'+i+'.csv'))
            print('\n>> SYSTEM MESSAGE >> all Days Are Processed Successfully!\n')            
            
            if app_type == 'all_in_one':
                print('>> SYSTEM MESSAGE >> Processing Live App Data!')
                for i in time_frames:
                    print('\n>> PROCESS MESSAGE >> Processing {} Timeframe ..'.format(i))
                    file_names = [f for f in listdir(join(data_dir, 'live_data', 'raw_data'))]
                    for file_name in file_names:
                        print('>> PROCESS MESSAGE >>     Processing {} Currency pair ..'.format(file_name[:6]))
                        df = pd.read_csv(join(data_dir, 'live_data', 'raw_data', file_name))
                        df = pd.DataFrame({'open':df[open_col].values,
                                           'high':df[high_col].values,
                                           'low':df[low_col].values,
                                           'close':df[close_col].values,
                                           'tick_volume':df[tick_vol_col].values}, index=pd.to_datetime(df[date_col], format=date_format))
                        idx = pd.date_range(df.index.date[0], freq='min', periods=1440)
                        df = df.reindex(idx, method='ffill')
                        df = df.replace(to_replace=np.nan, method='bfill')
                        df = process_ohlc(df=df, tf=i)
                        df.to_csv(join(data_dir, 'live_data', 'time_frames', i, file_name[:6]+'_'+i+'.csv'))

        #############################################################################################################################

        elif app_type == 'live_simulator':
            tf_list = listdir(join(data_dir, 'time_frames'))
            time_frames = []
            for i in range(1, len(tf_list)+1):
                for tf in tf_list:
                    if tf[1].isalpha():
                        x = int(tf[0])
                    else:
                        x = int(tf[:2])
                    if x == i:
                        time_frames.append(tf)
            print('>> SYSTEM MESSAGE >> Processing Live App Data!')
            for i in time_frames:
                print('\n>> PROCESS MESSAGE >> Processing {} Timeframe ..'.format(i))
                file_names = [f for f in listdir(join(data_dir, 'raw_data'))]
                for file_name in file_names:
                    print('>> PROCESS MESSAGE >>     Processing {} Currency pair ..'.format(file_name[:6]))
                    df = pd.read_csv(join(data_dir, 'raw_data', file_name))
                    df = pd.DataFrame({'open':df[open_col].values,
                                        'high':df[high_col].values,
                                        'low':df[low_col].values,
                                        'close':df[close_col].values,
                                        'tick_volume':df[tick_vol_col].values}, index=pd.to_datetime(df[date_col], format=date_format))
                    idx = pd.date_range(df.index.date[0], freq='min', periods=1440)
                    df = df.reindex(idx, method='ffill')
                    df = df.replace(to_replace=np.nan, method='bfill')
                    df = process_ohlc(df=df, tf=i)
                    df.to_csv(join(data_dir, 'time_frames', i, file_name[:6]+'_'+i+'.csv'))
    if save_logs:
        sys.stdout.close()
    return

if __name__ == '__main__':
    pass
    


__package_name__                 = 'FX-Manager'
__version__                      = '1.0.6'
__author__                       = 'Abdullah Bahi'
__author_email__                 = 'abdullahbahi@icloud.com'
__description__                  = 'Python package for developing and testing forex algorithmic trading strategies.'
__long_description_content_type__= "text/markdown"
__url__                          = 'https://github.com/AbdullahBahi/fxmanager'
__keywords__                     = ['forex', 'forex trading', 'algorethmic trading', 'backtesting']
__python_requires__              = '>=3.7'
__copyright__                    = "Copyright © 2021 Abdullah Bahi"
__credits__                      = ["Darwinex"]
__license__                      = "GNU General Public License (GPL)"
__maintainer__                   = "Abdullah Bahi"
__status__                       = "3 - Alpha"

__classifiers__      = ['Development Status :: 3 - Alpha',
                        'Intended Audience :: Developers',
                        'Topic :: Software Development :: Build Tools',
                        'License :: OSI Approved :: GNU General Public License (GPL)',
                        'Programming Language :: Python :: 3',
                        'Operating System :: OS Independent']

__scripts__          = ['scripts\\historical_simulation.py',
                        'scripts\\live_simulation.py', 
                        'scripts\\optimizer_test.py',
                        'scripts\\pp.py', 
                        'scripts\\setup.py', 
                        'scripts\\test_rates_historic.py']

__package_dir__      = {'':'src'}

__packages__         = ['fxmanager', 'fxmanager.basic', 'fxmanager.dwx', 'fxmanager.optimization', 'fxmanager.simulation', 'fxmanager.strategies']

__install_requires__ =['zmq',
                        'pandas>=0.25.1',
                        'numpy>=1.16.5',
                        'matplotlib>=3.1.1']




#!/usr/bin/env python

"""
This module contains a function that rans an equally weighted optimization test on historical data.

Functions:
    - run_test(): Runs an optimization test on historical data given an 'optimization_method' and 'optimization_objective'.
"""

import sys
from os import getcwd, listdir
from os.path import join
from time import sleep
from itertools import combinations
import pandas as pd 
import numpy as np
import matplotlib.pyplot as plt
import fxmanager.optimization.eq_weight_optimizer as optim
import fxmanager.basic.util as util
import fxmanager._metadata as md
from __main__ import __dict__

__dict__.update(md.__dict__)

def run_test(data_dir=None, optimization_method="", optimization_objective="", scenarios_per_day=100, risk_rewards=[], leverage=0.01, test_run=False, save_plots=False, save_logs=False):
    """
    Runs an optimization test given an 'optimization_method' and 'optimization_objective'.
    
    The test uses preprocessed historical data to construct a portfolio using data from day (n) and then we use this portfolio to calculate average expected returns in day (n+1). The expected returns are calculated for every single element in 'risk_rewards' list and then averaged over 'scenarios_per_day' every day.

    Args:
        - data_dir              : string with the directory or full path to the directory in which application data is kept. If the setup() function is used to create the recommended project structure, the default None value should be used.
        
        - optimization_method   : string with optimization method to be used. The supported optimization methods are:
            - mixed_timeframes : selects the optimal currency pair from each available timeframe, then selects the optimal equally weighted combination of these time frames.
            - single_timeframe : selects the optimal time frame from the available timeframes, then selects the optimal equally weighted combination of currency pairs in the selected time frame.
        
        - optimization_objective: string with optimization objective to be used. The supported optimization objectives are:
            - min_win_rate    : minimizes the required win rate for the portfolio to acheive a positive returns.
            - min_risk_reward : minimizes the required risk reward ratio required for the portfolio to acheive a positive returns.

        - scenarios_per_day     : integer indicating the number of scenarios over which the expected return is averaged every day. 
        - risk_rewards          : list of floats. the average expected return is calculated for each value in this list every day.
        - leverage              : float indicating the leverage. if no leverage is used set to 1.
        - test_run              : boolean flag used for testing purposes. If true, the constructed portfolios will consist of only one asset.
        - save_plots            : boolean flag. If True, the program will save result plots to 'data_dir\\visualizations\\' directory.
        - save_logs             : boolean flag. If True, the program will save result plots to 'data_dir\\logs\\equally_weighted_optimizer_logs.txt' file.
    Returns:
        - None
    
    Saved Results: test results are saved to 'data_dir\\visualizations' for plots and 'data_dir\\stats' for csv files.
        
        - Optimizer_Test_Win_Probs_&_Expexted_Rets_vs_Offsets.png : image with 2 sub-plots sharing the X-axis for the offsets, and Y axis for win probabilites and average expected returns.  
          Offsets are linearly separated floats added to the 'min_win_rate' argument, starting from (0.01, 0.02, ...), the value of 'min_win_rate' is incremented until the win probabilty becomes greater than or equal to 1.  
          the plot might also contain several lines per plot, each line represents calculations a single risk-reward ratio.
        
        - Optimizer_Test_Required_Win_Rates.png: image with a plot of constructed portfolios win rates on the Y-axis vs days on the X-axis.
        
        - Optimizer_Test_Risk _Reward_Ratios.png: image with a plot of constructed portfolios risk rewards on the Y-axis vs days on the X-axis.

        - Optimizer_Test_portfolios.csv : csv file with the constructed portfolios for every day. the index starts from 1 because every row represents the portfolio created using data from the day before so day 0 does not count.

        - Optimizer_Test_win_probs.csv : csv file with the data frame used for plotting the first sub-plot in 'Optimizer_Test_Win_Probs_&_Expexted_Rets_vs_Offsets.png'
        
        - Optimizer_Test_expexted_returns.csv : csv file with the data frame used for plotting the second sub-plot in 'Optimizer_Test_Win_Probs_&_Expexted_Rets_vs_Offsets.png'
    """

    if data_dir is None:
        data_dir = join(getcwd(), 'data')

    num_days = len(listdir(join(data_dir, 'time_frames')))
    tf_list = listdir(join(data_dir, 'time_frames', '0'))
    time_frames = []
    for i in range(1, len(tf_list)+1):
        for tf in tf_list:
            if tf[1].isalpha():
                x = int(tf[0])
            else:
                x = int(tf[:2])
            if x == i:
                time_frames.append(tf)
    currency_pairs = [x[:6] for x in listdir(join(data_dir, 'time_frames', '0', '1min'))]
    
    # Initialize data placeholders to empty dataframes
    offset_win_probs = pd.DataFrame()
    offset_er = pd.DataFrame()
    portfolios = pd.DataFrame()

    if save_logs:
        sys.stdout = open(join(data_dir, 'logs','equally_weighted_optimizer_logs.txt'), 'w')

    for d in range(num_days-1):
        print('-----------------------------------------------------------------------------------------------------------')
        print('-----------------------------------------------------------------------------------------------------------')
        print('                                             >> Processing Day {} <<'.format(d+1))
        print('-----------------------------------------------------------------------------------------------------------')
        print('-----------------------------------------------------------------------------------------------------------')
        ## STAGE 1 : Load best and worst returns
        print('\n>> SYSTEM MESSAGE >> STAGE 1: Loading Data ..\n')
        sleep(1)
        best_rets_df, wrst_rets_df = util.get_avg_rets(day=d, data_dir=data_dir, time_frames=time_frames, currency_pairs=currency_pairs)
        best_rets_df.index, wrst_rets_df.index = range(best_rets_df.shape[0]), range(best_rets_df.shape[0]) 

        ## STAGE 2 : Start Optimizing the portfolio
        print('>> SYSTEM MESSAGE >> Data Loaded Successfully!\n')
        print('>> SYSTEM MESSAGE >> STAGE 2: Optimizing Portfolio Parameters ..')
        
        optimized_portfolio_idxs, optimized_portfolio_tfs, optimized_portfolio_weights, \
        optimized_n, optimized_W_R, optimized_R_R = optim.optimize(best_rets_df,
                                                                    wrst_rets_df,
                                                                    method=optimization_method,
                                                                    objective=optimization_objective,
                                                                    test_run=test_run)
        optimized_portfolio_cps = [currency_pairs[x] for x in optimized_portfolio_idxs]

        ## STAGE 3 : Calculate the returns in the following day using the optimized portfolio
        print('\n>> SYSTEM MESSAGE >> Portfilio Construction is complete! ..\n')
        sleep(1)
        print('>> SYSTEM MESSAGE >> STAGE 3: Calculating Returns Using the Constructed Portfolio .. ')
        sleep(1)
        
        for rr in risk_rewards:
            win_probability = 0
            offset_W_R = 0
            offset_win_probs_day = pd.DataFrame()
            offset_er_day = pd.DataFrame()
            while win_probability < 1:
                if optimized_W_R+offset_W_R > 1:
                    break
                # print(offset_W_R, win_probability)
                er, win_probability = util.expected_return(day = d+1,
                                                            data_dir=data_dir,
                                                            time_frames = optimized_portfolio_tfs,
                                                            currency_pairs = optimized_portfolio_cps,
                                                            weights=optimized_portfolio_weights,
                                                            leverage=leverage,
                                                            win_rate = optimized_W_R + offset_W_R,
                                                            risk_reward = rr,
                                                            n_scenarios = scenarios_per_day)
                # break
                offset_win_probs_day[str(round(offset_W_R,2))] = [win_probability]
                offset_er_day[str(round(offset_W_R,2))] = [er]
                offset_W_R += 0.01
            # break
            offset_win_probs_day.index = [rr]
            offset_er_day.index = [rr]
            offset_win_probs = offset_win_probs.append(offset_win_probs_day, sort=True)
            offset_er = offset_er.append(offset_er_day, sort=True)

        ## STAGE 4 : Print the results
        print('\n>> SYSTEM MESSAGE >> Returns are Calculated Successfully!\n')
        print('>> SYSTEM MESSAGE >> STAGE 4: Printing Day results ..\n')
        sleep(1)
        print(f'\n>> PROCESS MESSAGE >> Day {d+1} Results: ')
        print(f'>> PROCESS MESSAGE >> Number of Assets Used in the Portfolio                : {optimized_n} Assets')
        print(f'>> PROCESS MESSAGE >> Portfolio Currency Pairs                              : {optimized_portfolio_cps}')
        print(f'>> PROCESS MESSAGE >> Portfolio Time Frames                                 : {optimized_portfolio_tfs}')
        print(f'>> PROCESS MESSAGE >> Portfolio Weights                                     : {optimized_portfolio_weights}')
        print(f'>> PROCESS MESSAGE >> Lowest Win Rate Required To Get Profit                : {round(optimized_W_R*100, 2)}%')
        print(f'>> PROCESS MESSAGE >> Portfolio Risk-Reward Ratio                           : {round(optimized_R_R*100, 2)}%')
        print(f'>> PROCESS MESSAGE >> Win Probability At {round(offset_W_R*100,2)}% Extra Win Rate                : {round(win_probability*100, 2)}%')
        print(f'>> PROCESS MESSAGE >> Average Expexted Return Across {scenarios_per_day} Scenarios          : {round(er*100, 2)}%')

        portfolios[d+1] = pd.Series({'optimized_W_R' : optimized_W_R,
                                     'optimized_R_R' : optimized_R_R,
                                     'optimized_n'   : optimized_n,
                                     'currency_pairs': optimized_portfolio_cps,
                                     'time_frames'    : optimized_portfolio_tfs,
                                     'weights': optimized_portfolio_weights
                                    })

    ## STAGE 5 : Save Results & Stats
    print('\n>> SYSTEM MESSAGE >> No More Days are Left! Go get More Data!\n')
    sleep(1)
    print('>> SYSTEM MESSAGE >> STAGE 4: Saving Results and Portfolios Summary..\n')
    sleep(1)

    if save_logs:
        sys.stdout.close()

    if save_plots:
        # Get average expected returns and win probs over all days for each risk reward value
        AVG_er = pd.DataFrame()
        AVG_win_probs = pd.DataFrame()
        for rr in risk_rewards:
            AVG_er[rr] = offset_er.loc[rr].mean(axis=0)
            AVG_win_probs[rr] = offset_win_probs.loc[rr].mean(axis=0)

        AVG_er.index.name, AVG_win_probs.index.name = 'Win Rate Offsets', 'Win Rate Offsets'

        # Create a figure with 2 subplots
        fig1, (ax1,ax2) = plt.subplots(2, 1, figsize = [10, 15])

        # Plot the Win Probabilites vs Win Rate Offsets in the upper subplot
        AVG_win_probs.plot(ax=ax1, grid=True, sharex=True, title='Win Probabilites & Expexted Returns vs Win Rate Offsets', legend=True)
        ax1.set_ylim(0,1)
        ax1.set_ylabel('Win Probabilites')
        ax1.legend(title="Risk Reward Ratios", fancybox=True)

        # Plot the Expexted Returns vs Win Rate Offsets in the lower subplot
        AVG_er.plot(ax=ax2, grid=True, sharex=True, legend=True)
        ax2.set_ylabel('Expexted Returns')
        ax2.legend(title="Risk Reward Ratios", fancybox=True)

        # Save the figure as png image
        fig1.savefig(join(data_dir,'visualizations','Optimizer_Test_Win_Probs_&_Expexted_Rets_vs_Offsets.png'))

        # Create a new figure
        fig2 = plt.figure(figsize = [15, 10])
        ax3 = plt.gca()

        # Plot the optimized win rates of all portfolios (x-axis is the days, one portfolio for each day)
        portfolios = portfolios.transpose()
        portfolios['optimized_W_R'].plot(ax=ax3, grid=True, title='Minimum Required Win Rates', label='Win Rates')
        ax3.axhline(y=portfolios['optimized_W_R'].mean(), color='r', linestyle='--', label='Average Win Rate')
        ax3.set_xlim(0,21)
        ax3.legend()
        ax3.set_ylabel('Win Rates')

        # Save the figure as png image
        fig2.savefig(join(data_dir,'visualizations','Optimizer_Test_Required_Win_Rates.png'))

        # Create a new figure
        fig3 = plt.figure(figsize = [15, 10])
        ax4 = plt.gca()

        # Plot the optimized risk reward ratios of all portfolios (x-axis is the days, one portfolio for each day)
        portfolios['optimized_R_R'].plot(ax=ax4, grid=True, title='Risk-Reward Ratios', label='Risk-Reward Ratios')
        ax4.axhline(y=portfolios['optimized_R_R'].mean(), color='r', linestyle='--', label='Average Risk-Reward Ratio')
        ax4.set_xlim(0,21)
        ax4.legend()
        ax4.set_ylabel('Risk-Reward ratios')

        # Save the figure as png image
        fig3.savefig(join(data_dir,'visualizations','Optimizer_Test_Risk _Reward_Ratios.png'))

    # portfolios = portfolios.transpose()
    portfolios.index.name, offset_win_probs.index.name, offset_er.index.name = 'days','days' ,'days'
    offset_er.index = np.array([np.arange(1, num_days)]*len(risk_rewards)).T.flatten()
    offset_win_probs.index = np.array([np.arange(1, num_days)]*len(risk_rewards)).T.flatten()
    portfolios.to_csv(join(data_dir, 'stats', 'Optimizer_Test_portfolios.csv'))
    offset_win_probs.to_csv(join(data_dir, 'stats', 'Optimizer_Test_win_probs.csv'))
    offset_er.to_csv(join(data_dir, 'stats', 'Optimizer_Test_expexted_returns.csv'))

    return




#!/usr/bin/env python

"""
This module contains helper functions used internally in this module and other modules in fxmanager's optimization sub-package. 

the functions defined in this module are designed for constructing optimal weighted portfolios. a portfolio consists of one or more assets where an asset in this context refers to a currency pair being traded in a specefic time frame.  
for example, if the currency pair 'EURUSD' is being traded in two different time frames, 5 and 10 minutes, where a new position is opened every 5 or 10 minutes, in this case the 'EURUSD_5min' and 'EURUSD_10min' are two different assets.

Public Functions:
    - optimize(): runs weighted portfolio optimization using data from a single day given an optimization method and objective.
        ** this function is designed to be used internally in 'fxmanager.optimization.eqw_optimization_test.run()' function. However, it can be used by the user separately to construct a portfolio for a given day without running the optimization test.
"""

from os import getcwd
from os.path import join
from itertools import combinations
from collections import deque
import pandas as pd
import numpy as np
import fxmanager.basic.util as util
import fxmanager._metadata as md
from __main__ import __dict__

__dict__.update(md.__dict__)

def mixed_timeframes_portfolio(bestN, wrstN, data_dir, objective='min_win_rate', print_progress=False):
    """
    Helper function for 'fxmanager.optimization.weight_optimizer.optimize()' function.

    selects the optimal currency pair from each timeframe in the given data, then selects the optimal weighted combination of these timeframes.
    """

    num_assets = bestN.shape[1]
    assert  (len(bestN) >= num_assets) & (len(wrstN) >= num_assets)

    rows = combinations(range(bestN.shape[0]), num_assets)
    cols = bestN.columns.values

    weights = util.get_weights(data_dir, num_assets=num_assets)
    weights = [np.array(x)/num_assets for x in weights]

    optimized_W_R = 100
    optimized_R_R = 100
    c = 0
    for c, row in enumerate(rows):
        for w in weights:
            for i in range(num_assets):
                w = deque(w)
                w.rotate(1)
                w = np.array(w)
                temp_df = bestN.loc[row, cols]
                diagonal = temp_df.to_numpy().diagonal()
                W_S_best = (w*diagonal).sum()
                temp_df = wrstN.loc[row, cols]
                diagonal = temp_df.to_numpy().diagonal()
                W_S_wrst = (w*diagonal).sum()
                
                W_R = (abs(W_S_wrst) / (W_S_best + abs(W_S_wrst)))
                R_R = abs(W_S_wrst) / W_S_best
        
                if objective == 'min_win_rate':
                    if W_R < optimized_W_R:
                        optimized_W_R = W_R
                        optimized_R_R = R_R
                        optimized_portfolio_idxs = row
                        optimized_portfolio_weights = w
                elif objective == 'min_risk_reward':
                    if R_R < optimized_R_R:
                        optimized_W_R = W_R
                        optimized_R_R = R_R
                        optimized_portfolio_idxs = row
                        optimized_portfolio_weights = w
                c += 1
                if print_progress:
                    if c%500 == 0:
                        print(f'>> PROCESS MESSAGE >> Finished: {round((c/ncr(bestN.shape[0], num_assets))*100, 2)}%     optimized_W_R: {round(optimized_W_R*100, 2)}%')
    if print_progress:
        print(f'>> PROCESS MESSAGE >> Finished: 100%     optimized_W_R: {round(optimized_W_R*100, 2)}%\n')
    
    return optimized_portfolio_idxs, cols, optimized_portfolio_weights, optimized_W_R, optimized_R_R

def single_timeframe_portfolio(best_tf, wrst_tf, data_dir, num_assets, objective='min_win_rate', print_progress=False):
    """
    Helper function for 'fxmanager.optimization.eq_weight_optimizer.optimize()' function.

    selects the optimal weighted combination of currency pairs in the given timeframe data.
    """

    assert  (len(best_tf) >= 1) & (len(wrst_tf) >= 1)

    rows = combinations(range(best_tf.shape[0]), num_assets)
    cols = best_tf.columns.values

    weights = util.get_weights(data_dir, num_assets=num_assets)
    weights = [np.array(x)/num_assets for x in weights]
    # print(weights)

    optimized_W_R = 100
    optimized_R_R = 100
    c = 0
    for c, row in enumerate(rows):
        for w in weights:
            for i in range(num_assets):
                w = deque(w)
                w.rotate(1)
                w = np.array(w).reshape(num_assets,1)
                W_S_best = (best_tf.loc[row, cols].to_numpy()*w).sum()
                W_S_wrst = (wrst_tf.loc[row, cols].to_numpy()*w).sum()
                
                W_R = (abs(W_S_wrst) / (W_S_best + abs(W_S_wrst)))
                R_R = abs(W_S_wrst) / W_S_best
        
                if objective == 'min_win_rate':
                    if W_R < optimized_W_R:
                        optimized_W_R = W_R
                        optimized_R_R = R_R
                        optimized_portfolio_idxs = row
                        optimized_portfolio_weights = w
                elif objective == 'min_risk_reward':
                    if R_R < optimized_R_R:
                        optimized_W_R = W_R
                        optimized_R_R = R_R
                        optimized_portfolio_idxs = row
                        optimized_portfolio_weights = w
                c += 1
                if print_progress:
                    if c%500 == 0:
                        print(f'>> PROCESS MESSAGE >> Finished: {round((c/ncr(best_tf.shape[0], num_assets))*100, 2)}%     optimized_W_R: {round(optimized_W_R*100, 2)}%')
    if print_progress:
        print(f'>> PROCESS MESSAGE >> Finished: 100%     optimized_W_R: {round(optimized_W_R*100, 2)}%\n')
    
    return optimized_portfolio_idxs, cols, optimized_portfolio_weights, optimized_W_R, optimized_R_R

def optimize(best_rets_df, wrst_rets_df, data_dir=None, method='mixed_timeframes', objective='min_win_rate', test_run=True):
    """
    runs weighted portfolio optimization using data from a single day given an optimization method and objective.

    Args:
        - best_rets_df: pandas dataframe of average best return for each asset, columns are time_frames and index is a range index.
        - wrst_rets_df: pandas dataframe of average worst return for each asset, columns are time_frames and index is a range index.
        - data_dir    : string with the directory name or full path to the directory in which application data is kept. If the setup()
                        function is used to create the recommended project structure, the default None value should be used.
        
        - method: string with optimization method to be used. The supported optimization methods are:
            - mixed_timeframes: selects the optimal currency pair from each available timeframe, then selects the optimal equally weighted
                                combination of these time frames.
            - single_timeframe: selects the optimal time frame from the available timeframes, then selects the optimal equally weighted
                                combination of currency pairs in the selected time frame.
        
        - objective: string with optimization objective to be used. The supported optimization objectives are:
            - min_win_rate   : minimizes the required win rate for the portfolio to acheive a positive returns.
            - min_risk_reward: minimizes the required risk reward ratio required for the portfolio to acheive a positive returns.
        
        - test_run: boolean flag used for testing purposes. If true, the constructed portfolios will consist of only one asset.

    Returns:
        - optimized_portfolio_idxs   : tuple with length equal to 'optimized_n' and values of integers that repesent the indices corresponding
                                       to the optimal currency pairs. 
        - optimized_portfolio_tfs    : numpy array with length of 'optimized_n' and values of optimal timeframes.
        - optimized_portfolio_weights: numpy array with length equal to 'optimized_n' and values of equal floats that sum up to 1.
        - optimized_n                : integer indicating the number of assets in the constructed portfolio.
        - optimized_W_R              : float indicating the minimum required win rate for the portfolio to acheive positive return.
        - optimized_R_R              : float indicating the minimum required risk reward for the portfolio to acheive positive return.

    ** this function is designed to be used internally in 'fxmanager.optimization.eqw_optimization_test.run()' function. However,
        it can be used by the user separately to construct a portfolio for a given day without running the optimization test.
    """

    if data_dir is None:
        data_dir = join(getcwd(), 'data')

    num_tfs = best_rets_df.shape[1]
    num_cps = best_rets_df.shape[0]
    
    if method == 'mixed_timeframes':
        optimized_W_R = 100 ## Use any large initial value
        optimized_R_R = 100
        
        for n in range(1,num_tfs+1): # 109293 total iterations
            print(f'>> PROCESS MESSAGE >> Processing for n = {n} .. \n')
            top_n_tfs, _, __ = util.top_n_timeframes(best_rets_df, wrst_rets_df, objective=objective, n=n)
            portfolio_idxs, portfolio_cols, portfolio_weights, W_R, R_R = mixed_timeframes_portfolio(best_rets_df[top_n_tfs],
                                                                                                    wrst_rets_df[top_n_tfs],
                                                                                                    data_dir = data_dir,
                                                                                                    objective = objective,
                                                                                                    print_progress = False)
            if objective == 'min_win_rate':
                if W_R < optimized_W_R:
                    optimized_W_R = W_R
                    optimized_R_R = R_R
                    optimized_portfolio_idxs = portfolio_idxs
                    optimized_portfolio_tfs = portfolio_cols
                    optimized_portfolio_weights = portfolio_weights
                    optimized_n = n
            elif objective == 'min_risk_reward':
                if R_R < optimized_R_R:
                    optimized_W_R = W_R
                    optimized_R_R = R_R
                    optimized_portfolio_idxs = portfolio_idxs
                    optimized_portfolio_tfs = portfolio_cols
                    optimized_portfolio_weights = portfolio_weights
                    optimized_n = n
            if test_run:
                break
    
    elif method == 'single_timeframe':
        optimized_W_R = 100 ## Use any large initial value
        optimized_R_R = 100
        top_n_tfs, _, __ = util.top_n_timeframes(best_rets_df, wrst_rets_df, objective=objective, n=1)
        
        for n in range(1,num_cps+1):
            print(f'>> PROCESS MESSAGE >> Processing for n = {n} .. \n')
            portfolio_idxs, portfolio_cols, portfolio_weights, W_R, R_R = single_timeframe_portfolio(best_tf = best_rets_df[top_n_tfs],
                                                                                                    wrst_tf = wrst_rets_df[top_n_tfs],
                                                                                                    data_dir = data_dir,
                                                                                                    num_assets = n,
                                                                                                    objective=objective,
                                                                                                    print_progress=False)
            if objective == 'min_win_rate':
                if W_R < optimized_W_R:
                    optimized_W_R = W_R
                    optimized_R_R = R_R
                    optimized_portfolio_idxs = portfolio_idxs
                    optimized_portfolio_tfs = portfolio_cols
                    optimized_portfolio_weights = portfolio_weights
                    optimized_n = n
            elif objective == 'min_risk_reward':
                if R_R < optimized_R_R:
                    optimized_W_R = W_R
                    optimized_R_R = R_R
                    optimized_portfolio_idxs = portfolio_idxs
                    optimized_portfolio_tfs = portfolio_cols
                    optimized_portfolio_weights = portfolio_weights
                    optimized_n = n
            if test_run:
                break

    return optimized_portfolio_idxs, optimized_portfolio_tfs, optimized_portfolio_weights, optimized_n, optimized_W_R, optimized_R_R

if __name__ == '__main__':
    data_dir = join(getcwd(), '..','datasets')
    time_frames = ['1min', '2min', '3min', '4min', '5min', '6min', '7min', '8min', '9min', '10min']
    currency_pairs = ['AUDUSD', 'EURUSD', 'GBPUSD', 'NZDUSD', 'USDCAD', 'USDCHF', 'USDCNH', 'USDDKK',
                  'USDHKD', 'USDHUF', 'USDJPY', 'USDMXN', 'USDNOK', 'USDPLN', 'USDSEK', 'USDSGD',
                  'USDZAR']
    d = 0
    best_rets_df, wrst_rets_df = get_rets(day=d, data_dir=data_dir, time_frames=time_frames, currency_pairs=currency_pairs)
    best_rets_df.index, wrst_rets_df.index = range(best_rets_df.shape[0]), range(best_rets_df.shape[0])
    
    optimized_portfolio_cps, optimized_portfolio_tfs, optimized_portfolio_weights, optimized_n, optimized_W_R, optimized_R_R = optimize(best_rets_df,
                                                                                                        wrst_rets_df,
                                                                                                        data_dir=data_dir,
                                                                                                        method='single_timeframe',
                                                                                                        objective='min_win_rate',
                                                                                                        test_run=False)
    print(f'Number of Assets Used in the Portfolio                : {optimized_n} Assets')
    print(f'Portfolio Currency Pairs                              : {optimized_portfolio_cps}')
    print(f'Portfolio Time Frames                                 : {optimized_portfolio_tfs}')
    print(f'Portfolio Weights                                     : {optimized_portfolio_weights}')
    print(f'Lowest Win Rate Required to get profit                : {round(optimized_W_R*100, 2)}%')
    print(f'Portfolio Risk-Reward ratio                           : {round(optimized_R_R*100, 2)}%')



# -*- coding: utf-8 -*-
"""
    DWX_ZMQ_Reporting.py
    --
    @author: Darwinex Labs (www.darwinex.com)
    
    Copyright (c) 2019 onwards, Darwinex. All rights reserved.
    
    Licensed under the BSD 3-Clause License, you may not use this file except 
    in compliance with the License. 
    
    You may obtain a copy of the License at:    
    https://opensource.org/licenses/BSD-3-Clause
"""

from pandas import DataFrame, to_datetime
from time import sleep

class DWX_ZMQ_Reporting():
    
    def __init__(self, _zmq):
        self._zmq = _zmq
        
    ##########################################################################
    
    def _get_open_trades_(self, _trader='Trader_SYMBOL', 
                          _delay=0.1, _wbreak=10):
        
        # Reset data output
        self._zmq._set_response_(None)
        
        # Get open trades from MetaTrader
        self._zmq._DWX_MTX_GET_ALL_OPEN_TRADES_()

        # While loop start time reference            
        _ws = to_datetime('now')
        
        # While data not received, sleep until timeout
        while self._zmq._valid_response_('zmq') == False:
            
            sleep(_delay)
            
            if (to_datetime('now') - _ws).total_seconds() > (_delay * _wbreak):
                break
        
        # If data received, return DataFrame
        if self._zmq._valid_response_('zmq'):
            
            _response = self._zmq._get_response_()
            
            if ('_trades' in _response.keys()
                and len(_response['_trades']) > 0):
                
                _df = DataFrame(data=_response['_trades'].values(),
                                index=_response['_trades'].keys())
                return _df[_df['_comment'] == _trader]
            
        # Default
        return DataFrame()
    
    ##########################################################################
    


# -*- coding: utf-8 -*-
"""
    DWX_ZMQ_Strategy.py
    --
    @author: Darwinex Labs (www.darwinex.com)
    
    Copyright (c) 2019 onwards, Darwinex. All rights reserved.
    
    Licensed under the BSD 3-Clause License, you may not use this file except 
    in compliance with the License. 
    
    You may obtain a copy of the License at:    
    https://opensource.org/licenses/BSD-3-Clause
"""

from fxmanager.dwx.DWX_ZeroMQ_Connector_v2_0_1_RC8 import DWX_ZeroMQ_Connector
from fxmanager.dwx.DWX_ZMQ_Execution import DWX_ZMQ_Execution
from fxmanager.dwx.DWX_ZMQ_Reporting import DWX_ZMQ_Reporting

class DWX_ZMQ_Strategy(object):
    
    def __init__(self, _name="DEFAULT_STRATEGY",    # Name 
                 _symbols=[('EURUSD',0.01),     # List of (Symbol,Lotsize) tuples
                           ('AUDNZD',0.01),
                           ('NDX',0.10),
                           ('UK100',0.1),
                           ('GDAXI',0.01),
                           ('XTIUSD',0.01),
                           ('SPX500',1.0),
                           ('STOXX50E',0.10),
                           ('XAUUSD',0.01)],
                 _broker_gmt=3,                 # Darwinex GMT offset
                 _pulldata_handlers = [],       # Handlers to process data received through PULL port.
                 _subdata_handlers = [],        # Handlers to process data received through SUB port.
                 _verbose=False):               # Print ZeroMQ messages
                 
        self._name = _name
        self._symbols = _symbols
        self._broker_gmt = _broker_gmt
        
        # Not entirely necessary here.
        self._zmq = DWX_ZeroMQ_Connector(_pulldata_handlers=_pulldata_handlers,
                                         _subdata_handlers=_subdata_handlers,
                                         _verbose=_verbose)
        
        # Modules
        self._execution = DWX_ZMQ_Execution(self._zmq)
        self._reporting = DWX_ZMQ_Reporting(self._zmq)
        
    ##########################################################################
    
    def _run_(self):
        
        """
        Enter strategy logic here
        """
         
    ##########################################################################



#!/usr/bin/env python

"""
This module contains 'Account' class which is used for creating a fully functional virtual forex trading accounts.

used for trading simulation and strategies backtesting. first the user creates an account instance with the required paramaters, then this
instance is passed to fxmanager's built-in historical or live simulators. 

Classes:
    - Account: Class that simulates forex trading accounts.
"""

from numpy import inf
import fxmanager._metadata as md
from __main__ import __dict__

__dict__.update(md.__dict__)

class Account():
    """
    Class that simulates forex trading accounts.
    
    Args:
        - balance         : float with initial balance in the account.
        - account_type    : string idicating account type, supported account types are
            - standard: lot size = 100,000
            - mini    : lot size = 10,000
            - micro   : lot size = 1000
            - nano    : lot size = 100
        - account_currency: string indicating the account currency (currency of initial balance).
        - leverage        : float indicating the leverage provided by the broker
        - volume_bounds   : tuple of 2 floats, minimum and maximum availabe trading volumes per one position.

    Public Methods:
        - open_pisition() : opens a new position and updates the state of the account accordingly.
        - close_position(): closes an opened position and updates the state of the account accordingly.
        - update()        : updates the state of the account (including oppened positions) with the current market prices.
    """

    def __init__(self, balance = 100000.0, account_type = 'standard', account_currency = 'usd', leverage = 0.01, volume_bounds = (0.01, 8.0)):
        ## constants
        LOT_SIZES = {"standard":100000.0, "mini":10000.0, "micro":1000.0, "nano":100.0}
        self._LOT_SIZE = LOT_SIZES[account_type.strip().lower()]
        self._LEVERAGE = leverage
        self._ACCOUNT_CURRENCY = account_currency.strip().lower()
        self._MIN_VOLUME = volume_bounds[0]
        self._MAX_VOLUME = volume_bounds[1]

        self._balance = balance ## account balance (updated every time a position is closed)
        self._equity = balance ## balance + closed_profit(updated every time a position is closed)
        self._live_equity = balance ## balance + live_profit (updated every time unit using close price)
        self._margin = 0.0 ## required deposit to hold current positionss (updated every time a position is opened  or closed)
        self._free_margin = balance ## equity - margin (updated every time a position is opened  or closed)
        self._live_free_margin = balance ## live equity - margin (updated every time unit using close price)
        self._margin_level = inf ## equity / margin (updated every time a position is opened  or closed)
        self._live_margin_level = inf ## live equity / margin (updated every time unit using close price)
        self._profit = 0.0 ## final profit (updated every time a position is closed)
        self._live_profit = 0.0 ## current profit  (updated every time unit using close price)
        # self._look_back_counter = {'AUDUSD':0, 'EURUSD':0, 'GBPUSD':0, 'NZDUSD':0, 'USDCAD':0, 'USDCHF':0, 'USDCHN':0, 'USDDKK':0, 'USDHKD':0, 'USDHUF':0, 'USDJPY':0, 'USDMXN':0, 'USDNOK':0, 'USDPLN':0, 'USDSEK':0, 'USDSGD':0, 'USDZAR':0}
        self._positions = {} ## dictionary to hold current positions {"Ticket":{..Position properties}, ..} (updated every time a position is opened or closed)
    
    def _get_lot_value(self, price, lot_size):
        return price * lot_size
    
    def _get_margin(self, lot_value, leverage, volume):
        return lot_value * leverage * volume
    
    def _get_profit(self, open_lot_value, close_lot_value, volume, order_type):
        order_types = {"sell":-1.0, "buy":1.0} ## must set profit to (close - open)
        return (close_lot_value - open_lot_value) * volume * order_types[order_type.strip().lower()]
    
    def open_pisition(self, ticket, base_currency, quote_currency, time_frame, weight, SL, TP, order_type, prices, period, order_idx, risk_factor=0.8):
        """
        opens a new position and updates the state of the account accordingly.
        """
        
        if not(base_currency.lower() != self._ACCOUNT_CURRENCY and quote_currency.lower() != self._ACCOUNT_CURRENCY):
            # FIRST GET OPEN PRICE
            price_types = {"sell":"bid", "buy":"ask"}
            open_price = prices[price_types[order_type]]
            if base_currency.lower() == self._ACCOUNT_CURRENCY:
                open_price = 1 / open_price

            # if price_types[order_type] == "buy":
            #     inc = 0.0001
            # else:
            #     inc = -0.0001
            # open_price += inc
            
            # NOW CALCULATE THE VOLUME
            # if self._look_back_counter[base_currency+quote_currency] < look_back:
            #     volume = 0.01
            #     self._look_back_counter[base_currency+quote_currency] +=1
            # else:
            max_margin = weight * self._balance ## maximum margin requirment that can be covered
            volume = (max_margin*risk_factor) / (open_price * self._LOT_SIZE * self._LEVERAGE)
            
            if volume < self._MIN_VOLUME:
                return False
            elif volume > self._MAX_VOLUME:
                volume = self._MAX_VOLUME
            else:
                pass

            # NOW UPDATE PORTFOLIO STATE
            ## 1 ## Margin
            open_lot_value = self._get_lot_value(open_price, self._LOT_SIZE)
            margin = self._get_margin(lot_value = open_lot_value, leverage = self._LEVERAGE, volume = volume)
            
            if self._balance < margin: ## if the current balance is less than the margin requirement >> don't open the position
                return False
            
            self._margin += margin

            ## 2 ## Free Margin
            self._free_margin = self._equity - self._margin

            ## 3 ## Margin Level
            if round(self._margin,2) == 0:
                self._margin_level = inf
            else:
                self._margin_level = round(self._equity,4) / round(self._margin,4)

            ## 4 ## Positions
            self._positions[ticket] = {
                'ticket': ticket,
                'order_idx': order_idx,
                'order_type': order_type,
                'volume': volume,
                'base_currency': base_currency,
                'quote_currency': quote_currency,
                'time_frame': time_frame,
                'open_price': open_price,
                'margin': margin,
                'SL': open_price - (SL * 0.001 * open_price), ## Formula: open_price - (SL * Maximum expected pip difference * open_price)
                'TP': open_price + (TP * 0.001 * open_price), ## Formula: open_price + (TP * Maximum expected pip difference * open_price)
                'live_profit': 0,
                'period': period,
                'weight': weight
            }
        
        else:
            print("\n>> PROCESS MESSAGE >> Position profit cannot be calculated! Currency pair must contain account currency.\n")
            return False
        return True
        

    def close_position(self, ticket, prices):
        """
        closes an opened position and updates the state of the account accordingly.
        """

        # FIRST GET POSITION PARAMETERS
        order_type = self._positions[ticket]['order_type']
        volume = self._positions[ticket]['volume']
        base_currency = self._positions[ticket]['base_currency']
        quote_currency = self._positions[ticket]['quote_currency']
        open_price = self._positions[ticket]['open_price']
        SL = self._positions[ticket]['SL']
        TP = self._positions[ticket]['TP']
        
        # NOW GET CLOSE PRICE
        price_types = {"sell":"ask", "buy":"bid"}
        close_price = prices[price_types[order_type]]
        if base_currency.lower() == self._ACCOUNT_CURRENCY:
            close_price = 1 / close_price

        # if price_types[order_type] == "buy":
        #     inc = -0.0001
        # else:
        #     inc = 0.0001
        # open_price += inc

        # NOW UPDATE PORTFOLIO STATE
        ## 1 ## Profit
        open_lot_value = self._get_lot_value(open_price, self._LOT_SIZE)
        close_lot_value = self._get_lot_value(close_price, self._LOT_SIZE)
        position_profit = self._get_profit(open_lot_value, close_lot_value, volume, order_type)
        self._profit += position_profit

        ## 2 ## balance
        self._balance += position_profit

        ## 3 ## Equity
        self._equity = self._balance

        ## 4 ## Margin
        margin = self._positions[ticket]['margin']
        self._margin -= margin

        ## 5 ## Free Margin
        self._free_margin = self._equity - self._margin

        ## 6 ## Margin Level
        if round(self._margin,2) == 0:
            self._margin_level = inf
        else:
            self._margin_level = round(self._equity,4) / round(self._margin,4)
        
        ## 7 ## Positions
        self._positions.pop(ticket)

        return position_profit, margin, close_price
    
    def update(self, prices):
        """
        updates the state of the account (including oppened positions) with the current market prices.
        """

        # UPDATE PORTFOLIO STATE
        ## 1 ## Live Profit & Periods
        self._live_profit = 0.0
        for i in self._positions:
            # Update Periods
            if self._positions[i]['period'] > 0:
                self._positions[i]['period'] -= 1

            # Get required position parameters
            order_type = self._positions[i]['order_type']
            volume = self._positions[i]['volume']
            base_currency = self._positions[i]['base_currency']
            quote_currency = self._positions[i]['quote_currency']
            cp = base_currency + quote_currency
            open_price = self._positions[i]['open_price']
            
            # Get current price
            price_types = {"sell":"ask", "buy":"bid"} ## note that bid and ask are reversed because it's a CLOSE price
            current_price = prices.iloc[-1][[cp+'_ask_close', cp+'_bid_close']]
            current_price.index = ['ask', 'bid']
            current_price = current_price[price_types[order_type]]
            
            if base_currency.lower() == self._ACCOUNT_CURRENCY:
                current_price = 1 / current_price

            # Calculate current profit
            open_lot_value = self._get_lot_value(open_price, self._LOT_SIZE)
            current_lot_value = self._get_lot_value(current_price, self._LOT_SIZE) ## base/quote lot value at if position is closed at current price
            live_profit = self._get_profit(open_lot_value, current_lot_value, volume, order_type)
            self._positions[i]['live_profit'] = live_profit
            self._live_profit += live_profit

        ## 2 ## Live Equity
        self._live_equity = self._balance + self._live_profit

        ## 3 ## Live Free Margin
        self._live_free_margin = self._live_equity - self._margin

        ## 4 ## Live Margin Level
        self._live_margin_level = self._live_equity / self._margin





"""
This sub-package contains the modules used for historical and live trading sumulations.

Modules:
    - historic: This module contains helper functions used internally in this module and a public function to run historical simulation.
    
    - live    : This module contains helper functions used internally in this module and a public function to run live simulation with live price feed from MT4.
"""

from . import historic
from . import live
import fxmanager._metadata as md
from __main__ import __dict__

__dict__.update(md.__dict__)


#!/usr/bin/env python

"""
This script is used for pre-processing raw market prices data to be compaitable with FX-Manager.

This script should be used after running 'setup.py' script. data is organized in the following
structure:
    >> 'data\\raw_data\\':
        >> '0\\':
            >> '1min\\':
                >> 'currency_pair_raw_data_file.csv'
                >> 'other_currency_pair_raw_data_file.csv'
                >> ...
            >> '2min\\':
                >> 'currency_pair_raw_data_file.csv'
                >> 'other_currency_pair_raw_data_file.csv'
                >> ...
            >> ...
        >> '1\\':
            >> ...
        >> ...

Usage:
      fx_pp [options]

Options:
    -h, --help                 : show this help message and exit.
    
    -v, --version              : show the version of FX-Manager and exit.
    
    -at, --app_type            : Supported App types: 
                                    - back_tester          : if you want to test trading strategies 
                                                             histoical data on FX-Manager virtual account.
                                    - portfolio_optimizer  : if you want to asses different portfolio optimization
                                                             methods and objectives using histoical data.
                                    - live_simulator       : if you want to test trading strategies in real time
                                                             using MT4 Live data on FX-Manager virtual account.
                                    - all_in_one           : if you want to use all the previous features.
    
    -rdt, --raw_data_type      : Supported Raw Data types: 
                                    - daily_bid_ask  : Bid/Ask ticks in daily seperate files.
                                      (a file for each currency_pair in each day)
                                    - daily_candles  : 1-min candles (OHLCV) in daily seperate files.
                                      (a file for each currency_pair in each day)
                                    - period_candles : 1-min candles (OHLCV) over a period.
                                      (a file for each currency_pair for all the days)
    
    -sl, --save_logs           : boolean flag, if True, the program logs are saved to
                                 'data\\logs\\preprocessing_logs.txt' file
"""

doc = __doc__

from os import getcwd
from os.path import join
import json
import argparse
from fxmanager.basic.util import preprocess
import fxmanager._metadata as md
from __main__ import __dict__

__dict__.update(md.__dict__)

# Add Optional arguments
parser = argparse.ArgumentParser(add_help=False)
parser.add_argument('-h','--help', default=False, action='store_true')
parser.add_argument('-v','--version', action='version', version=__version__)
parser.add_argument('-at','--app_type', type=str, default = 'all_in_one')
parser.add_argument('-rdt','--raw_data_type', type=str, default = 'daily_candles')
parser.add_argument('-sl','--save_logs', default=False, action='store_true')

# Parse arguments
in_args = parser.parse_args()

if in_args.help:
    import sys
    print(doc)
    sys.exit()
    
with open('raw_data_formats.json') as f:
    raw_data_formats = json.load(f)

if in_args.raw_data_type == 'daily_bid_ask':
    raw_data_format = raw_data_formats['daily_bid_ask']
elif in_args.raw_data_type == 'daily_candles':
    raw_data_format = raw_data_formats['daily_candles']
elif in_args.raw_data_type == 'period_candles':
    raw_data_format = raw_data_formats['period_candles']
else:
    pass 

preprocess(app_type = in_args.app_type,
            raw_data_type = in_args.raw_data_type,
            raw_data_format = raw_data_format,
            save_logs = in_args.save_logs)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This Module is used as an API to get historic data from the MT4 EA.

This is a modified version of an example of using the Darwinex ZeroMQ Connector for Python 3 and MetaTrader 4 PULL REQUEST for v2.0.1 in which a Client requests rate history from a Daily from a start date to an end date. The user creates a 'rates_historic' object and using the methods 'get_daily_candles' and 'get_period_candles' the user client can get the historic data of the desired symbols. 

The original example is referenced `here <https://github.com/darwinex/dwx-zeromq-connector/blob/master/v2.0.1/python/examples/template/strategies/rates_historic.py>`_.

Through commmand HIST, this client can select multiple rates from an INSTRUMENT (symbol, timeframe).
For example, to receive rates from instruments EURUSD(M1), between two dates, it will send this command to the Server, through its PUSH channel:

"HIST;EURUSD;1;2019.01.04 00:00:00;2019.01.14 00:00:00"
  
Original Author: `raulMrello <https://www.linkedin.com/in/raul-martin-19254530/>`_

Modified by    : `AbdullahBahi <https://www.linkedin.com/in/abdullahbahi/>`_
"""

#############################################################################
# DWX-ZMQ required imports 
#############################################################################

# Import ZMQ-Strategy from relative path
from fxmanager.dwx.DWX_ZMQ_Strategy import DWX_ZMQ_Strategy

#############################################################################
# Other required imports
#############################################################################

import pandas as pd
from threading import Thread, Lock
from time import sleep
from os.path import join
import fxmanager._metadata as md
from __main__ import __dict__

__dict__.update(md.__dict__)

#############################################################################
# Class derived from DWZ_ZMQ_Strategy includes data processor for PULL,SUB data
#############################################################################

class rates_historic(DWX_ZMQ_Strategy):
    
    def __init__(self, 
                 _name="PRICES_SUBSCRIPTIONS",
                 _delay=0.1,
                 _broker_gmt=3,
                 _verbose=False):
        
        # call DWX_ZMQ_Strategy constructor and passes itself as data processor for handling
        # received data on PULL and SUB ports 
        super().__init__(_name,
                         [],          # Empty symbol list (not needed for this example)
                         _broker_gmt,
                         [self],      # Registers itself as handler of pull data via self.onPullData()
                         [self],      # Registers itself as handler of sub data via self.onSubData()
                         _verbose)
        
        # This strategy's variables
        self._delay = _delay
        self._verbose = _verbose
        self._finished = False

        # lock for acquire/release of ZeroMQ connector
        self._lock = Lock()
        
    ##########################################################################    
    def isFinished(self):        
        """ Check if execution finished"""
        return self._finished
        
    ##########################################################################    
    def onPullData(self, data):        
        """
        Callback to process new data received through the PULL port
        """        
        # print responses to request commands
        # print('>> DWX MESSAGE >> Historic from ExpertAdvisor={}'.format(data))
        # _topic, _msg = data.split(" ")
        # print(_msg)
        pass
        
    ##########################################################################    
    def onSubData(self, data):        
        """
        Callback to process new data received through the SUB port
        """
        # split msg to get topic and message
        _topic, _msg = data.split(" ")
        # print('>> DWX MESSAGE >> Data on Topic={} with Message={}'.format(_topic, _msg))
        
    ##########################################################################    
    def get_daily_candles(self, save_to='', currency_pair='', date=''):        
        """
        Request historic data and returns data of type 'daily_candles'
        """        
        self._finished = False
                
        # request rates
        print(f'>> DWX MESSAGE >> Requesting {currency_pair} Rates from {date}')
        self._zmq._DWX_MTX_SEND_HIST_REQUEST_(_symbol=currency_pair,
                                              _timeframe=1,
                                              _start=date + ' 00:00:00',
                                              _end=date + ' 23:59:00')
        while (currency_pair + '_M1') not in self._zmq._History_DB.keys():
          sleep(1)
        print('>> DWX MESSAGE >> Saving Data to CSV file ...\n')        
        df = pd.DataFrame(self._zmq._History_DB[currency_pair + '_M1'])
        df.to_csv(join(save_to, currency_pair + '_' + date + '.csv'))
        del df
        # finishes (removes all subscriptions)  
        self.stop()
  
    def get_period_candles(self, save_to='', currency_pair='', start='', end=''):        
        """
        Request historic data and returns data of type 'period_candles'
        """        
        self._finished = False
                
        # request rates
        print(f'>> DWX MESSAGE >> Requesting {currency_pair} Rates from {start} to {end}')
        self._zmq._DWX_MTX_SEND_HIST_REQUEST_(_symbol=currency_pair,
                                              _timeframe=1,
                                              _start=start + ' 00:00:00',
                                              _end=end + ' 23:59:00')
        while (currency_pair + '_M1') not in self._zmq._History_DB.keys():
          sleep(5)
        print('>> DWX MESSAGE >> Saving Data to CSV file ...\n')  
        df = pd.DataFrame(self._zmq._History_DB[currency_pair + '_M1'])
        df.to_csv(join(save_to, currency_pair + '_' + start + '_' + end + '.csv'))
        del df
        
        # finishes (removes all subscriptions)  
        self.stop()

    ##########################################################################    
    def stop(self):
      """
      unsubscribe from all market symbols and exits
      """
      # remove subscriptions and stop symbols price feeding
      try:
        # Acquire lock
        self._lock.acquire()
        self._zmq._DWX_MTX_UNSUBSCRIBE_ALL_MARKETDATA_REQUESTS_()
        # print('>> DWX MESSAGE >> Unsubscribing from all topics')
          
      finally:
        # Release lock
        self._lock.release()
        sleep(self._delay)
      
      self._finished = True

""" -----------------------------------------------------------------------------------------------
    -----------------------------------------------------------------------------------------------
    SCRIPT SETUP
    -----------------------------------------------------------------------------------------------
    -----------------------------------------------------------------------------------------------
"""
if __name__ == "__main__":
  
  # Define timeframes
  timeframes = {'M1': 1,
                'M5': 5,
                'M15': 15,
                'M30': 30,
                'H1': 60,
                'H4': 240,
                'D1': 1440,
                'W1': 10080,
                'MN1': 43200
                }
  dates = {'M1': '2021.06.17 00:00:00',
          'M5': '2021.06.12 00:00:00',
          'M15': '2021.06.03 00:00:00',
          'M30': '2021.05.18 00:00:00',
          'H1': '2021.04.18 00:00:00',
          'H4': '2020.11.18 00:00:00',
          'D1': '2017.06.18 00:00:00',
          'W1': '2011.06.18 00:00:00',
          'MN1': '2001.06.18 00:00:00'
          }
  
  ## Recommended periods for different timeframes
  # > M1:  1 day       (2021.06.17 00:00:00 >> 2021.06.18 23:57:59)
  # > M5:  5 days      (2021.06.12 00:00:00 >> 2021.06.18 23:57:59)
  # > M15: 15 days     (2021.06.03 00:00:00 >> 2021.06.18 23:57:59)
  # > M30: 1 month     (2021.05.18 00:00:00 >> 2021.06.18 23:57:59)
  # > H1:  2 months    (2021.04.18 00:00:00 >> 2021.06.18 23:57:59)
  # > H4:  8 months    (2020.11.18 00:00:00 >> 2021.06.18 23:57:59)
  # > D1:  4 years     (2017.06.18 00:00:00 >> 2021.06.18 23:57:59)
  # > W1:  10 years    (2011.06.18 00:00:00 >> 2021.06.18 23:57:59)
  # > MN:  20 years    (2001.06.18 00:00:00 >> 2021.06.18 23:57:59)
  timeframe = 'M15'
  start = dates[timeframe]
  end = '2021.06.18 23:57:59'
  currency_pair = 'NZDUSD'
  
  # creates object with a predefined configuration: historic EURGBP_D1 between 4th adn 14th January 2019
  print('Loading example...')
  example = rates_historic()  

  # Starts example execution
  print('Running example...')

  example.run(currency_pair = currency_pair, timeframe = timeframes[timeframe], tf=timeframe, start=start, end=end)

  # Waits example termination
  print('Waiting example termination...')
  while not example.isFinished():
    sleep(1)
  print('Bye!!!')
  sleep(1)



#!/usr/bin/env python

"""
This script is used for setting up a new FX-Manager project.

Usage:
      create_fx_proj [options]

Options:
      -h, --help                : show this help message and exit.
      
      -v, --version             : show the version of FX-Manager and exit.

      -at, --app_type           : Supported App types: 
                                    - back_tester          : if you want to test trading strategies using histoical data
                                                             on FX-Manager virtual account.
                                    - portfolio_optimizer  : if you want to asses different portfolio optimization methods
                                                             and objectives using histoical data.
                                    - live_simulator       : if you want to test trading strategies in real time using MT4
                                                             Live data on FX-Manager virtual account.
                                    - all_in_one           : if you want to use all the previous features.

      -rdd, --raw_data_dir       : path to the directory where raw historical data files exist.
                                   path could be relative or full.
      
      -rdt, --raw_data_type      : Supported Raw Data types: 
                                    - daily_bid_ask  : Bid/Ask ticks in daily seperate files.
                                      (a file for each currency_pair in each day)
                                    - daily_candles  : 1-min candles in daily seperate files.
                                      (a file for each currency_pair in each day)
                                    - period_candles : 1-min candles over a period.
                                      (a file for each currency_pair for all the days)
      
      -lrdd, --live_raw_data_dir : path to the directory where raw historical data files used for
                                   live simulation exist. path could be relative or full.
      
      -fnf, --file_name_format   : string of white spaces with the same length as raw data file names.
                                   Givin that each file name has information about date and currency pair
                                   symbol, the white spaces in the same location as these information are
                                   replaced as follows:
                                    - day, month and year in the date are replaced with (dd), (mm), (yyyy) respectively.
                                    - currency pair symbol is replaced with (cccccc).
                                    Example:
                                    if the file name is "AUDUSD_Ticks_02.08.2021-02.08.2021.csv"
                                    then the -fnf is:
                                                        "cccccc       dd mm yyyy               "
      
      -nd, --num_days           : total number of work days in the dataset.
      -ncp, --num_currency_pairs : number of currency pairs in the dataset.
      -ntf, --num_time_frames    : number of timeframes in the dataset.
"""

doc = __doc__ 

import argparse
from fxmanager.basic.util import setup
import fxmanager._metadata as md
from __main__ import __dict__

__dict__.update(md.__dict__)

# Add Optional arguments
parser = argparse.ArgumentParser(add_help=False)
parser.add_argument('-h','--help', default=False, action='store_true')
parser.add_argument('-v','--version', action='version', version=__version__)
parser.add_argument('-at','--app_type', type=str, default = 'all_in_one')
parser.add_argument('-rdd','--raw_data_dir', type=str, default = 'raw_data')
parser.add_argument('-rdt','--raw_data_type', type=str, default = 'daily_candles')
parser.add_argument('-lrdd','--live_raw_data_dir', type=str, default = 'live_raw_data')    
parser.add_argument('-fnf','--file_name_format', type=str, default = 'cccccc yyyy mm dd    ')
parser.add_argument('-nd','--num_days', type=int, default = 2)
parser.add_argument('-ncp','--num_currency_pairs', type=int, default = 1)
parser.add_argument('-ntf','--num_time_frames', type=int, default = 10)

# Parse arguments
in_args = parser.parse_args()

if in_args.help:
  import sys
  print(doc)
  sys.exit()

# File Name Formats:
#     - daily_bid_ask  : 'cccccc       dd mm yyyy               '
#     - daily_candles  : 'cccccc yyyy mm dd    '
#     - period_candles : 'cccccc yyyy mm dd               '

setup(app_type = in_args.app_type,
        raw_data_dir = in_args.raw_data_dir,
        live_raw_data_dir = in_args.live_raw_data_dir,
        raw_data_type = in_args.raw_data_type,
        file_name_format = in_args.file_name_format,
        num_days = in_args.num_days,
        num_cps = in_args.num_currency_pairs,
        num_tfs = in_args.num_time_frames)


# -*- coding: utf-8 -*-
"""
    DWX_ZeroMQ_Connector_v2_0_1_RC8.py
    --
    @author: Darwinex Labs (www.darwinex.com)
    
    Copyright (c) 2017-2019, Darwinex. All rights reserved.
    
    Licensed under the BSD 3-Clause License, you may not use this file except 
    in compliance with the License. 
    
    You may obtain a copy of the License at:    
    https://opensource.org/licenses/BSD-3-Clause
"""

import zmq
from time import sleep
from pandas import DataFrame, Timestamp
from threading import Thread

# 30-07-2019 10:58 CEST
from zmq.utils.monitor import recv_monitor_message

class DWX_ZeroMQ_Connector():

    """
    Setup ZeroMQ -> MetaTrader Connector
    """
    def __init__(self, 
                 _ClientID='dwx-zeromq',    # Unique ID for this client
                 _host='localhost',         # Host to connect to
                 _protocol='tcp',           # Connection protocol
                 _PUSH_PORT=32768,          # Port for Sending commands
                 _PULL_PORT=32769,          # Port for Receiving responses
                 _SUB_PORT=32770,           # Port for Subscribing for prices
                 _delimiter=';',
                 _pulldata_handlers = [],   # Handlers to process data received through PULL port.
                 _subdata_handlers = [],    # Handlers to process data received through SUB port.
                 _verbose=True,             # String delimiter
                 _poll_timeout=1000,        # ZMQ Poller Timeout (ms)
                 _sleep_delay=0.001,        # 1 ms for time.sleep()
                 _monitor=False):           # Experimental ZeroMQ Socket Monitoring
    
        ######################################################################
        
        # Strategy Status (if this is False, ZeroMQ will not listen for data)
        self._ACTIVE = True
        
        # Client ID
        self._ClientID = _ClientID
        
        # ZeroMQ Host
        self._host = _host
        
        # Connection Protocol
        self._protocol = _protocol

        # ZeroMQ Context
        self._ZMQ_CONTEXT = zmq.Context()
        
        # TCP Connection URL Template
        self._URL = self._protocol + "://" + self._host + ":"
        
        # Handlers for received data (pull and sub ports)
        self._pulldata_handlers = _pulldata_handlers
        self._subdata_handlers = _subdata_handlers

        # Ports for PUSH, PULL and SUB sockets respectively
        self._PUSH_PORT = _PUSH_PORT
        self._PULL_PORT = _PULL_PORT
        self._SUB_PORT = _SUB_PORT
        
        # Create Sockets
        self._PUSH_SOCKET = self._ZMQ_CONTEXT.socket(zmq.PUSH)
        self._PUSH_SOCKET.setsockopt(zmq.SNDHWM, 1)        
        self._PUSH_SOCKET_STATUS = {'state': True, 'latest_event': 'N/A'}
        
        self._PULL_SOCKET = self._ZMQ_CONTEXT.socket(zmq.PULL)
        self._PULL_SOCKET.setsockopt(zmq.RCVHWM, 1)
        self._PULL_SOCKET_STATUS = {'state': True, 'latest_event': 'N/A'}
        
        self._SUB_SOCKET = self._ZMQ_CONTEXT.socket(zmq.SUB)
        
        # Bind PUSH Socket to send commands to MetaTrader
        self._PUSH_SOCKET.connect(self._URL + str(self._PUSH_PORT))
        print("[INIT] Ready to send commands to METATRADER (PUSH): " + str(self._PUSH_PORT))
        
        # Connect PULL Socket to receive command responses from MetaTrader
        self._PULL_SOCKET.connect(self._URL + str(self._PULL_PORT))
        print("[INIT] Listening for responses from METATRADER (PULL): " + str(self._PULL_PORT))
        
        # Connect SUB Socket to receive market data from MetaTrader
        print("[INIT] Listening for market data from METATRADER (SUB): " + str(self._SUB_PORT))
        self._SUB_SOCKET.connect(self._URL + str(self._SUB_PORT))
        
        # Initialize POLL set and register PULL and SUB sockets
        self._poller = zmq.Poller()
        self._poller.register(self._PULL_SOCKET, zmq.POLLIN)
        self._poller.register(self._SUB_SOCKET, zmq.POLLIN)
        
        # Start listening for responses to commands and new market data
        self._string_delimiter = _delimiter
        
        # BID/ASK Market Data Subscription Threads ({SYMBOL: Thread})
        self._MarketData_Thread = None
        
        # Socket Monitor Threads
        self._PUSH_Monitor_Thread = None
        self._PULL_Monitor_Thread = None
        
        # Market Data Dictionary by Symbol (holds tick data)
        self._Market_Data_DB = {}   # {SYMBOL: {TIMESTAMP: (BID, ASK)}}
        
        # History Data Dictionary by Symbol (holds historic data of the last HIST request for each symbol)
        self._History_DB = {}   # {SYMBOL_TF: [{'time': TIME, 'open': OPEN_PRICE, 'high': HIGH_PRICE, 
                                #               'low': LOW_PRICE, 'close': CLOSE_PRICE, 'tick_volume': TICK_VOLUME, 
                                #               'spread': SPREAD, 'real_volume': REAL_VOLUME}, ...]}
                                
        # Temporary Order STRUCT for convenience wrappers later.
        self.temp_order_dict = self._generate_default_order_dict()
        
        # Thread returns the most recently received DATA block here
        self._thread_data_output = None
        
        # Verbosity
        self._verbose = _verbose
        
        # ZMQ Poller Timeout
        self._poll_timeout = _poll_timeout
        
        # Global Sleep Delay
        self._sleep_delay = _sleep_delay
        
        # Begin polling for PULL / SUB data
        self._MarketData_Thread = Thread(target=self._DWX_ZMQ_Poll_Data_, 
                                         args=(self._string_delimiter,
                                               self._poll_timeout,))
        self._MarketData_Thread.daemon = True
        self._MarketData_Thread.start()
        
        ###########################################
        # Enable/Disable ZeroMQ Socket Monitoring #
        ###########################################
        if _monitor == True:
            
            # ZeroMQ Monitor Event Map
            self._MONITOR_EVENT_MAP = {}
            
            print("\n[KERNEL] Retrieving ZeroMQ Monitor Event Names:\n")
            
            for name in dir(zmq):
                if name.startswith('EVENT_'):
                    value = getattr(zmq, name)
                    print(f"{value}\t\t:\t{name}")
                    self._MONITOR_EVENT_MAP[value] = name
            
            print("\n[KERNEL] Socket Monitoring Config -> DONE!\n")
        
            # Disable PUSH/PULL sockets and let MONITOR events control them.
            self._PUSH_SOCKET_STATUS['state'] = False
            self._PULL_SOCKET_STATUS['state'] = False
            
            # PUSH
            self._PUSH_Monitor_Thread = Thread(target=self._DWX_ZMQ_EVENT_MONITOR_, 
                                               args=("PUSH",
                                                     self._PUSH_SOCKET.get_monitor_socket(),))
            
            self._PUSH_Monitor_Thread.daemon = True
            self._PUSH_Monitor_Thread.start()
            
            # PULL
            self._PULL_Monitor_Thread = Thread(target=self._DWX_ZMQ_EVENT_MONITOR_, 
                                               args=("PULL",
                                                     self._PULL_SOCKET.get_monitor_socket(),))
            
            self._PULL_Monitor_Thread.daemon = True
            self._PULL_Monitor_Thread.start()
       
    ##########################################################################
    
    def _DWX_ZMQ_SHUTDOWN_(self):
        
        # Set INACTIVE
        self._ACTIVE = False
        
        # Get all threads to shutdown
        if self._MarketData_Thread is not None:
            self._MarketData_Thread.join()
            
        if self._PUSH_Monitor_Thread is not None:
            self._PUSH_Monitor_Thread.join()
            
        if self._PULL_Monitor_Thread is not None:            
            self._PULL_Monitor_Thread.join()
        
        # Unregister sockets from Poller
        self._poller.unregister(self._PULL_SOCKET)
        self._poller.unregister(self._SUB_SOCKET)
        print("\n++ [KERNEL] Sockets unregistered from ZMQ Poller()! ++")
        
        # Terminate context 
        self._ZMQ_CONTEXT.destroy(0)
        print("\n++ [KERNEL] ZeroMQ Context Terminated.. shut down safely complete! :)")
        
    ##########################################################################
    
    """
    Set Status (to enable/disable strategy manually)
    """
    def _setStatus(self, _new_status=False):
    
        self._ACTIVE = _new_status
        print("\n**\n[KERNEL] Setting Status to {} - Deactivating Threads.. please wait a bit.\n**".format(_new_status))
                
    ##########################################################################
    
    """
    Function to send commands to MetaTrader (PUSH)
    """
    def remote_send(self, _socket, _data):
        
        if self._PUSH_SOCKET_STATUS['state'] == True:
            try:
                _socket.send_string(_data, zmq.DONTWAIT)
            except zmq.error.Again:
                print("\nResource timeout.. please try again.")
                sleep(self._sleep_delay)
        else:
            print('\n[KERNEL] NO HANDSHAKE ON PUSH SOCKET.. Cannot SEND data')
      
    ##########################################################################
    
    def _get_response_(self):
        return self._thread_data_output
    
    ##########################################################################
    
    def _set_response_(self, _resp=None):
        self._thread_data_output = _resp
    
    ##########################################################################
    
    def _valid_response_(self, _input='zmq'):
        
        # Valid data types
        _types = (dict,DataFrame)
        
        # If _input = 'zmq', assume self._zmq._thread_data_output
        if isinstance(_input, str) and _input == 'zmq':
            return isinstance(self._get_response_(), _types)
        else:
            return isinstance(_input, _types)
            
        # Default
        return False
    
    ##########################################################################
    
    """
    Function to retrieve data from MetaTrader (PULL)
    """
    def remote_recv(self, _socket):
        
        if self._PULL_SOCKET_STATUS['state'] == True:
            try:
                msg = _socket.recv_string(zmq.DONTWAIT)
                return msg
            except zmq.error.Again:
                print("\nResource timeout.. please try again.")
                sleep(self._sleep_delay)
        else:
            print('\r[KERNEL] NO HANDSHAKE ON PULL SOCKET.. Cannot READ data', end='', flush=True)
            
        return None
        
    ##########################################################################
    
    # Convenience functions to permit easy trading via underlying functions.
    
    # OPEN ORDER
    def _DWX_MTX_NEW_TRADE_(self, _order=None):
        
        if _order is None:
            _order = self._generate_default_order_dict()
        
        # Execute
        self._DWX_MTX_SEND_COMMAND_(**_order)
        
    # MODIFY ORDER
    # _SL and _TP given in points. _price is only used for pending orders. 
    def _DWX_MTX_MODIFY_TRADE_BY_TICKET_(self, _ticket, _SL, _TP, _price=0):
        
        try:
            self.temp_order_dict['_action'] = 'MODIFY'
            self.temp_order_dict['_ticket'] = _ticket
            self.temp_order_dict['_SL'] = _SL
            self.temp_order_dict['_TP'] = _TP
            self.temp_order_dict['_price'] = _price
            
            # Execute
            self._DWX_MTX_SEND_COMMAND_(**self.temp_order_dict)
            
        except KeyError:
            print("[ERROR] Order Ticket {} not found!".format(_ticket))
    
    # CLOSE ORDER
    def _DWX_MTX_CLOSE_TRADE_BY_TICKET_(self, _ticket):
        
        try:
            self.temp_order_dict['_action'] = 'CLOSE'
            self.temp_order_dict['_ticket'] = _ticket
            
            # Execute
            self._DWX_MTX_SEND_COMMAND_(**self.temp_order_dict)
            
        except KeyError:
            print("[ERROR] Order Ticket {} not found!".format(_ticket))
            
    # CLOSE PARTIAL
    def _DWX_MTX_CLOSE_PARTIAL_BY_TICKET_(self, _ticket, _lots):
        
        try:
            self.temp_order_dict['_action'] = 'CLOSE_PARTIAL'
            self.temp_order_dict['_ticket'] = _ticket
            self.temp_order_dict['_lots'] = _lots
            
            # Execute
            self._DWX_MTX_SEND_COMMAND_(**self.temp_order_dict)
            
        except KeyError:
            print("[ERROR] Order Ticket {} not found!".format(_ticket))
            
    # CLOSE MAGIC
    def _DWX_MTX_CLOSE_TRADES_BY_MAGIC_(self, _magic):
        
        try:
            self.temp_order_dict['_action'] = 'CLOSE_MAGIC'
            self.temp_order_dict['_magic'] = _magic
            
            # Execute
            self._DWX_MTX_SEND_COMMAND_(**self.temp_order_dict)
            
        except KeyError:
            pass
    
    # CLOSE ALL TRADES
    def _DWX_MTX_CLOSE_ALL_TRADES_(self):
        
        try:
            self.temp_order_dict['_action'] = 'CLOSE_ALL'
            
            # Execute
            self._DWX_MTX_SEND_COMMAND_(**self.temp_order_dict)
            
        except KeyError:
            pass
        
    # GET OPEN TRADES
    def _DWX_MTX_GET_ALL_OPEN_TRADES_(self):
        
        try:
            self.temp_order_dict['_action'] = 'GET_OPEN_TRADES'
                        
            # Execute
            self._DWX_MTX_SEND_COMMAND_(**self.temp_order_dict)
            
        except KeyError:
            pass
    
    # DEFAULT ORDER DICT
    def _generate_default_order_dict(self):
        return({'_action': 'OPEN',
                  '_type': 0,
                  '_symbol': 'EURUSD',
                  '_price': 0.0,
                  '_SL': 500, # SL/TP in POINTS, not pips.
                  '_TP': 500,
                  '_comment': self._ClientID,
                  '_lots': 0.01,
                  '_magic': 123456,
                  '_ticket': 0})
        
    ##########################################################################

    """
    Function to construct messages for sending HIST commands to MetaTrader

    Because of broker GMT offset _end time might have to be modified.
    """
    def _DWX_MTX_SEND_HIST_REQUEST_(self,
                                 _symbol='EURUSD',
                                 _timeframe=1440,
                                 _start='2020.01.01 00:00:00',
                                 _end=Timestamp.now().strftime('%Y.%m.%d %H:%M:00')):
                                 #_end='2019.01.04 17:05:00'):
        
        _msg = "{};{};{};{};{}".format('HIST',
                                     _symbol,
                                     _timeframe,
                                     _start,
                                     _end)

        # Send via PUSH Socket
        self.remote_send(self._PUSH_SOCKET, _msg)
    
    
    ##########################################################################
    """
    Function to construct messages for sending TRACK_PRICES commands to 
    MetaTrader for real-time price updates
    """
    def _DWX_MTX_SEND_TRACKPRICES_REQUEST_(self,
                                 _symbols=['EURUSD']):
        _msg = 'TRACK_PRICES'
        for s in _symbols:
          _msg = _msg + ";{}".format(s)

        # Send via PUSH Socket
        self.remote_send(self._PUSH_SOCKET, _msg)
    
    
    ##########################################################################
    """
    Function to construct messages for sending TRACK_RATES commands to 
    MetaTrader for OHLC
    """
    def _DWX_MTX_SEND_TRACKRATES_REQUEST_(self,
                                 _instruments=[('EURUSD_M1', 'EURUSD',1)]):
        _msg = 'TRACK_RATES'                                 
        for i in _instruments:
          _msg = _msg + ";{};{}".format(i[1], i[2])
          
        # Send via PUSH Socket
        self.remote_send(self._PUSH_SOCKET, _msg)
    
    
    ##########################################################################
    
    
    ##########################################################################
    """
    Function to construct messages for sending Trade commands to MetaTrader
    """
    def _DWX_MTX_SEND_COMMAND_(self, _action='OPEN', _type=0,
                                 _symbol='EURUSD', _price=0.0,
                                 _SL=50, _TP=50, _comment="Python-to-MT",
                                 _lots=0.01, _magic=123456, _ticket=0):
        
        _msg = "{};{};{};{};{};{};{};{};{};{};{}".format('TRADE',_action,_type,
                                                         _symbol,_price,
                                                         _SL,_TP,_comment,
                                                         _lots,_magic,
                                                         _ticket)
        
        # Send via PUSH Socket
        self.remote_send(self._PUSH_SOCKET, _msg)
        
        """
         compArray[0] = TRADE or DATA
         compArray[1] = ACTION (e.g. OPEN, MODIFY, CLOSE)
         compArray[2] = TYPE (e.g. OP_BUY, OP_SELL, etc - only used when ACTION=OPEN)
         
         For compArray[0] == DATA, format is: 
             DATA|SYMBOL|TIMEFRAME|START_DATETIME|END_DATETIME
         
         // ORDER TYPES: 
         // https://docs.mql4.com/constants/tradingconstants/orderproperties
         
         // OP_BUY = 0
         // OP_SELL = 1
         // OP_BUYLIMIT = 2
         // OP_SELLLIMIT = 3
         // OP_BUYSTOP = 4
         // OP_SELLSTOP = 5
         
         compArray[3] = Symbol (e.g. EURUSD, etc.)
         compArray[4] = Open/Close Price (ignored if ACTION = MODIFY)
         compArray[5] = SL
         compArray[6] = TP
         compArray[7] = Trade Comment
         compArray[8] = Lots
         compArray[9] = Magic Number
         compArray[10] = Ticket Number (MODIFY/CLOSE)
         """
        # pass
    
    ##########################################################################
    
    """
    Function to check Poller for new reponses (PULL) and market data (SUB)
    """
    
    def _DWX_ZMQ_Poll_Data_(self, 
                           string_delimiter=';',
                           poll_timeout=1000):
        
        while self._ACTIVE:
            
            sleep(self._sleep_delay) # poll timeout is in ms, sleep() is s.
            
            sockets = dict(self._poller.poll(poll_timeout))
            
            # Process response to commands sent to MetaTrader
            if self._PULL_SOCKET in sockets and sockets[self._PULL_SOCKET] == zmq.POLLIN:
                
                if self._PULL_SOCKET_STATUS['state'] == True:
                    try:
                        
                        # msg = self._PULL_SOCKET.recv_string(zmq.DONTWAIT)
                        msg = self.remote_recv(self._PULL_SOCKET)
                        
                        # If data is returned, store as pandas Series
                        if msg != '' and msg != None:
                            
                            try:
                                _data = eval(msg)
                                if '_action' in _data and _data['_action'] == 'HIST':
                                    _symbol = _data['_symbol']
                                    if '_data' in _data.keys():
                                        if _symbol not in self._History_DB.keys():
                                            self._History_DB[_symbol] = {}
                                        self._History_DB[_symbol] = _data['_data']
                                    else:
                                        print('No data found. MT4 often needs multiple requests when accessing data of symbols without open charts.')
                                        print('message: ' + msg)
                                
                                # invokes data handlers on pull port
                                for hnd in self._pulldata_handlers:
                                    hnd.onPullData(_data)
                                
                                self._thread_data_output = _data
                                if self._verbose:
                                    print(_data) # default logic
                                    
                            except Exception as ex:
                                _exstr = "Exception Type {0}. Args:\n{1!r}"
                                _msg = _exstr.format(type(ex).__name__, ex.args)
                                print(_msg)
                   
                    except zmq.error.Again:
                        pass # resource temporarily unavailable, nothing to print
                    except ValueError:
                        pass # No data returned, passing iteration.
                    except UnboundLocalError:
                        pass # _symbol may sometimes get referenced before being assigned.
                
                else:
                    print('\r[KERNEL] NO HANDSHAKE on PULL SOCKET.. Cannot READ data.', end='', flush=True)
            
            # Receive new market data from MetaTrader
            if self._SUB_SOCKET in sockets and sockets[self._SUB_SOCKET] == zmq.POLLIN:
                
                try:
                    msg = self._SUB_SOCKET.recv_string(zmq.DONTWAIT)
                    
                    if msg != "":

                        _timestamp = str(Timestamp.now('UTC'))[:-6]
                        _symbol, _data = msg.split(" ")
                        if len(_data.split(string_delimiter)) == 2:
                            _bid, _ask = _data.split(string_delimiter)   
                                                                   
                        
                            if self._verbose:
                                print("\n[" + _symbol + "] " + _timestamp + " (" + _bid + "/" + _ask + ") BID/ASK")                    
                    
                            # Update Market Data DB
                            if _symbol not in self._Market_Data_DB.keys():
                                self._Market_Data_DB[_symbol] = {}
                            
                            self._Market_Data_DB[_symbol][_timestamp] = (float(_bid), float(_ask))

                        elif len(_data.split(string_delimiter)) == 8:
                            _time, _open, _high, _low, _close, _tick_vol, _spread, _real_vol = _data.split(string_delimiter)
                            if self._verbose:
                                print("\n[" + _symbol + "] " + _timestamp + " (" + _time + "/" + _open + "/" + _high + "/" + _low + "/" + _close + "/" + _tick_vol + "/" + _spread + "/" + _real_vol + ") TIME/OPEN/HIGH/LOW/CLOSE/TICKVOL/SPREAD/VOLUME")                    
                            # Update Market Rate DB
                            if _symbol not in self._Market_Data_DB.keys():
                                self._Market_Data_DB[_symbol] = {}
                            self._Market_Data_DB[_symbol][_timestamp] = (int(_time), float(_open), float(_high), float(_low), float(_close), int(_tick_vol), int(_spread), int(_real_vol))

                        # invokes data handlers on sub port
                        for hnd in self._subdata_handlers:
                            hnd.onSubData(msg)

                except zmq.error.Again:
                    pass # resource temporarily unavailable, nothing to print
                except ValueError:
                    pass # No data returned, passing iteration.
                except UnboundLocalError:
                    pass # _symbol may sometimes get referenced before being assigned.
                    
        print("\n++ [KERNEL] _DWX_ZMQ_Poll_Data_() Signing Out ++")
                
    ##########################################################################
    
    """
    Function to subscribe to given Symbol's BID/ASK feed from MetaTrader
    """
    def _DWX_MTX_SUBSCRIBE_MARKETDATA_(self, 
                                       _symbol='EURUSD'):
        
        # Subscribe to SYMBOL first.
        self._SUB_SOCKET.setsockopt_string(zmq.SUBSCRIBE, _symbol)
        
        print("[KERNEL] Subscribed to {} BID/ASK updates. See self._Market_Data_DB.".format(_symbol))
    
    """
    Function to unsubscribe to given Symbol's BID/ASK feed from MetaTrader
    """
    def _DWX_MTX_UNSUBSCRIBE_MARKETDATA_(self, _symbol):
        
        self._SUB_SOCKET.setsockopt_string(zmq.UNSUBSCRIBE, _symbol)
        print("\n**\n[KERNEL] Unsubscribing from " + _symbol + "\n**\n")
        
        
    """
    Function to unsubscribe from ALL MetaTrader Symbols
    """
    def _DWX_MTX_UNSUBSCRIBE_ALL_MARKETDATA_REQUESTS_(self):
        
        # 31-07-2019 12:22 CEST
        for _symbol in self._Market_Data_DB.keys():
            self._DWX_MTX_UNSUBSCRIBE_MARKETDATA_(_symbol=_symbol)
        
    ##########################################################################
    
    def _DWX_ZMQ_EVENT_MONITOR_(self, 
                                socket_name, 
                                monitor_socket):
        
        # 05-08-2019 11:21 CEST
        while self._ACTIVE:
            
            sleep(self._sleep_delay) # poll timeout is in ms, sleep() is s.
            
            # while monitor_socket.poll():
            while monitor_socket.poll(self._poll_timeout):
                
                try:
                    evt = recv_monitor_message(monitor_socket, zmq.DONTWAIT)
                    evt.update({'description': self._MONITOR_EVENT_MAP[evt['event']]})
                    
                    # print(f"\r[{socket_name} Socket] >> {evt['description']}", end='', flush=True)
                    print(f"\n[{socket_name} Socket] >> {evt['description']}")
                    
                    # Set socket status on HANDSHAKE
                    if evt['event'] == 4096:        # EVENT_HANDSHAKE_SUCCEEDED
                        
                        if socket_name == "PUSH":
                            self._PUSH_SOCKET_STATUS['state'] = True
                            self._PUSH_SOCKET_STATUS['latest_event'] = 'EVENT_HANDSHAKE_SUCCEEDED'
                            
                        elif socket_name == "PULL":
                            self._PULL_SOCKET_STATUS['state'] = True
                            self._PULL_SOCKET_STATUS['latest_event'] = 'EVENT_HANDSHAKE_SUCCEEDED'
                            
                        # print(f"\n[{socket_name} Socket] >> ..ready for action!\n")
                            
                    else:    
                        # Update 'latest_event'
                        if socket_name == "PUSH":
                            self._PUSH_SOCKET_STATUS['state'] = False
                            self._PUSH_SOCKET_STATUS['latest_event'] = evt['description']
                            
                        elif socket_name == "PULL":
                            self._PULL_SOCKET_STATUS['state'] = False
                            self._PULL_SOCKET_STATUS['latest_event'] = evt['description']
                
                    if evt['event'] == zmq.EVENT_MONITOR_STOPPED:
                        
                        # Reinitialize the socket
                        if socket_name == "PUSH":
                            monitor_socket = self._PUSH_SOCKET.get_monitor_socket()
                        elif socket_name == "PULL":
                            monitor_socket = self._PULL_SOCKET.get_monitor_socket()
                        
                except Exception as ex:
                    _exstr = "Exception Type {0}. Args:\n{1!r}"
                    _msg = _exstr.format(type(ex).__name__, ex.args)
                    print(_msg)
               
        # Close Monitor Socket
        monitor_socket.close()
        
        print(f"\n++ [KERNEL] {socket_name} _DWX_ZMQ_EVENT_MONITOR_() Signing Out ++")
            
    ##########################################################################
    
    def _DWX_ZMQ_HEARTBEAT_(self):
        self.remote_send(self._PUSH_SOCKET, "HEARTBEAT;")
        
    ##########################################################################

##############################################################################

def _DWX_ZMQ_CLEANUP_(_name='DWX_ZeroMQ_Connector',
                      _globals=globals(), 
                      _locals=locals()):
    
    print('\n++ [KERNEL] Initializing ZeroMQ Cleanup.. if nothing appears below, no cleanup is necessary, otherwise please wait..')
    try:
        _class = _globals[_name]
        _locals = list(_locals.items())
        
        for _func, _instance in _locals:
            if isinstance(_instance, _class): 
                print(f'\n++ [KERNEL] Found & Destroying {_func} object before __init__()')
                eval(_func)._DWX_ZMQ_SHUTDOWN_()
                print('\n++ [KERNEL] Cleanup Complete -> OK to initialize DWX_ZeroMQ_Connector if NETSTAT diagnostics == True. ++\n')
           
    except Exception as ex:
        
        _exstr = "Exception Type {0}. Args:\n{1!r}"
        _msg = _exstr.format(type(ex).__name__, ex.args)
            
        if 'KeyError' in _msg:
            print('\n++ [KERNEL] Cleanup Complete -> OK to initialize DWX_ZeroMQ_Connector. ++\n')
        else:
            print(_msg)
            
##############################################################################



"""
FX-Manager is a python package for developing and testing forex algorithmic trading strategies.

Main Features:
  1. historical data collection and pre-processing: 
      - with FX-Manager, you can easily collect historical data from MT4 history center and preprocess it in just a few lines of code.

  2. flexibility with historical data formats:
      - whether the data is OHLCV or bid-ask, whether it's daily or over a period of time, all what it takes is a simple configuration and the data will be ready to be used in FX-Managers applications.

  3. open virtual trading account with no limitations:
      - most forex and stock brokers offer virtual trading accounts for new traders to learn trading with no risk of losing money. this might be sufficient for manual trading, but when it comes to algorithmic trading, it's not practical to use broker's virtual accounts to test trading strategies due to their limitations in terms of balance and time it takes to open a new account when an account is exhausted.
      - with FX-Manager, there is no need to be constrained with broker's limitations, instead of performing actual trading in your broker's account each time a trading strategy ia tested, FX-Manager creates an instance of a virtual account with all properties as an actual account and performs trading simulation on this instance without affecting the actual account on MT4 trading platform.
      - user can initialize this instance with any amount of balance and any account type with no limitations.

  4. run real-time trading simulation using real market prices:
      - using FX-Manager's virtual accounts, you can test your user-defined trading strategy by runinig real-time trading simulation with real market prices from MT4 live price feed.
      - this allows you to run multiple simulations in the same time because each simulation uses it's own account instance without affecting the account connected with MT4. MT4 only acts as a price feed source for all running instances.
      - the simulation shows account state every time period that is defined by user.

  5. run historical trading simulation using historical market prices:
      - in addition to real-time testing, you can also test your trading strategies using historical data and preview results in a very nice and well-stractured manner.

  6. biult-in portfolio construction and optimization:
      - what currency pairs to invest in and what time frame should be used for each currency pair are both the most important questions for any forex trader.
      - with FX-Manager portfolio construction feature. you can construct optimized portfolios consisting of diversified collection of currency pairs in the best time frame for each pair. these portfolios are constructed with historical data and can be used for historical and live trading simulations.

Sub-Packages:
  1. basic        : contains mudules of helper functions used intrnally in all FX-Manager's sub-packages and functions used for setting up new FX-Manager projects and preprocessing data.
  2. dwx          : contains a modified version of Darwinex ZeroMQ connector which is used for connecting client code to MT4 trading platform.
  3. optimization : contains the mudules used for portfolio construction and optimization.
  4. simulation   : contains the modules used for historical and live trading sumulations.
  5. strategies   : contains the mudules used for building and injecting trading strategies int FX-Manager's trading simulations.
"""

from . import basic
from . import dwx
from . import optimization
from . import simulation
from . import strategies
import fxmanager._metadata as md
from __main__ import __dict__

__dict__.update(md.__dict__)


"""
This sub-package contains mudules of helper functions used intrnally in all FX-Manager's sub-packages.

in addition to the helper functions, the sub-package contains functions used for setting up new FX-Manager projects and preprocessing data to be compaitable with FX-Manager. it also contains the module used for simulating the trading accounts.

Modules:
    - util    : This module contains helper functions used internally in other fxmanager sub-packages. it also includes public functions to be used by the user to setup fxmanager project structure and pre-process the data.

    - account : This module contains 'Account' class which is used for creating a fully functional virtual forex trading accounts.
"""

from . import util
from . import account
import fxmanager._metadata as md
from __main__ import __dict__

__dict__.update(md.__dict__)


"""
This sub-package contains the mudules used for building and injecting trading strategies int FX-Manager's trading simulations.

Public Modules:
    - naieve_momentum: This module contains simple implementation of momentum trading strategy, used as default for simulations.
"""

from . import naieve_momentum
from . import template
import fxmanager._metadata as md
from __main__ import __dict__

__dict__.update(md.__dict__)


"""
This sub-package contains the mudules used for portfolio construction and optimization.

Modules:
    - eq_weight_optimizer   : This module contains helper functions used internally in this module and other  modules in fxmanager's optimization sub-package. 
    
    - eqw_optimization_test : This module contains a function that rans an equally weighted optimization test on historical data.

    - weight_optimizer      : This module contains a function that rans a weighted optimization test on historical data.

    - w_optimization_test   : This module contains helper functions used internally in this module and other modules in fxmanager's optimization sub-package. 
"""

from . import eq_weight_optimizer
from . import eqw_optimization_test
from . import w_optimization_test
from . import weight_optimizer
import fxmanager._metadata as md
from __main__ import __dict__

__dict__.update(md.__dict__)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""    
This is a modified version of an example of using the Darwinex ZeroMQ Connector for Python 3 and MetaTrader 4 PULL REQUEST.

The original example is referenced `here <https://github.com/darwinex/dwx-zeromq-connector/blob/master/v2.0.1/python/examples/template/strategies/prices_subscriptions.py>`_, This Module is used as an API to get real time data from the MT4 EA. The user creates a 'prices_subscriptions' object with the desired symbols list in the client applicaction, then the real-time price feed can be accessed from the self._recent_prices dictionary.


Through commmand TRACK_PRICES, this client can select multiple SYMBOLS for price tracking.
For example, to receive real-time bid-ask prices from symbols EURUSD and GDAXI, this client will send this command to the Server, through its PUSH channel:

"TRACK_PRICES;EURUSD;GDAXI"

Server will answer through the PULL channel with a json response like this:

{'_action':'TRACK_PRICES', '_data': {'symbol_count':2}}

or if errors, then: 

{'_action':'TRACK_PRICES', '_data': {'_response':'NOT_AVAILABLE'}}

Once subscribed to this feed, it will receive through the SUB channel, prices in this format:

"EURUSD BID;ASK"
    
--
  
Original Author: `raulMrello <https://www.linkedin.com/in/raul-martin-19254530/>`_

Modified by    : `AbdullahBahi <https://www.linkedin.com/in/abdullahbahi/>`_
"""

#############################################################################
# DWX-ZMQ required imports 
#############################################################################

# Import ZMQ-Strategy from relative path
from fxmanager.dwx.DWX_ZMQ_Strategy import DWX_ZMQ_Strategy

#############################################################################
# Other required imports
#############################################################################

from threading import Thread, Lock
from time import sleep
import fxmanager._metadata as md
from __main__ import __dict__

__dict__.update(md.__dict__)

#############################################################################
# Class derived from DWZ_ZMQ_Strategy includes data processor for PULL,SUB data
#############################################################################

class prices_subscriptions(DWX_ZMQ_Strategy):
    
    def __init__(self, 
                 _name="PRICES_SUBSCRIPTIONS",
                 _symbols=['EURUSD','GDAXI'],
                 _delay=0.1,
                 _broker_gmt=3,
                 _verbose=False):
        
        # call DWX_ZMQ_Strategy constructor and passes itself as data processor for handling
        # received data on PULL and SUB ports 
        super().__init__(_name,
                         _symbols,
                         _broker_gmt,
                         [self],      # Registers itself as handler of pull data via self.onPullData()
                         [self],      # Registers itself as handler of sub data via self.onSubData()
                         _verbose)
        
        # This strategy's variables
        self._symbols = _symbols
        self._delay = _delay
        self._verbose = _verbose
        self._finished = False
        self._data_available = False
        self._recent_prices = {'AUDUSD':0, 'EURUSD':0, 'GBPUSD':0, 'NZDUSD':0, 'USDCAD':0, 'USDCHF':0,
                               'USDCHN':0, 'USDDKK':0, 'USDHKD':0, 'USDHUF':0, 'USDJPY':0, 'USDMXN':0,
                               'USDNOK':0, 'USDPLN':0, 'USDSEK':0, 'USDSGD':0, 'USDZAR':0, 'BTCUSD':0,
                               'USDCZK':0, 'USDTRY':0, 'EURAUD':0, 'EURCAD':0, 'EURCHF':0, 'EURCZK':0,
                               'EURDKK':0, 'EURGBP':0, 'EURHKD':0, 'EURHUF':0, 'EURJPY':0, 'EURMXN':0,
                               'EURNOK':0, 'EURNZD':0, 'EURPLN':0, 'EURSEK':0, 'EURSGD':0, 'EURTRY':0,
                               'EURZAR':0}

        # lock for acquire/release of ZeroMQ connector
        self._lock = Lock()
        
    ##########################################################################    
    def isFinished(self):        
        """ Check if execution finished"""
        return self._finished
        
    ##########################################################################    
    def onPullData(self, data):        
        """
        Callback to process new data received through the PULL port
        """        
        # print responses to request commands
        print('>> DWX MESSAGE >> Response from ExpertAdvisor={}'.format(data))
        
    ##########################################################################    
    def onSubData(self, data):        
        """
        Callback to process new data received through the SUB port
        """
        self._data_available = True
        # split msg to get topic and message
        _topic, _msg = data.split(" ")
        self._recent_prices[_topic] = [float(i) for i in _msg.strip().split(';')]

    ##########################################################################    
    def run(self):        
        """
        Starts price subscriptions
        """        
        self._finished = False

        # Subscribe to all symbols in self._symbols to receive bid,ask prices
        self.__subscribe_to_price_feeds()

    ##########################################################################    
    def stop(self):
      """
      unsubscribe from all market symbols and exits
      """
        
      # remove subscriptions and stop symbols price feeding
      try:
        # Acquire lock
        self._lock.acquire()
        self._zmq._DWX_MTX_UNSUBSCRIBE_ALL_MARKETDATA_REQUESTS_()
        print('>> DWX MESSAGE >> Unsubscribing from all topics')
          
      finally:
        # Release lock
        self._lock.release()
        sleep(self._delay)
      
      try:
        # Acquire lock
        self._lock.acquire()
        self._zmq._DWX_MTX_SEND_TRACKPRICES_REQUEST_([])        
        print('>> DWX MESSAGE >> Removing symbols list')
        sleep(self._delay)
        self._zmq._DWX_MTX_SEND_TRACKRATES_REQUEST_([])
        print('>> DWX MESSAGE >> Removing instruments list')

      finally:
        # Release lock
        self._lock.release()
        sleep(self._delay)

      self._finished = True


    ##########################################################################
    def __subscribe_to_price_feeds(self):
      """
      Starts the subscription to the self._symbols list setup during construction.
      1) Setup symbols in Expert Advisor through self._zmq._DWX_MTX_SUBSCRIBE_MARKETDATA_
      2) Starts price feeding through self._zmq._DWX_MTX_SEND_TRACKPRICES_REQUEST_
      """
      if len(self._symbols) > 0:
        # subscribe to all symbols price feeds
        for _symbol in self._symbols:
          try:
            # Acquire lock
            self._lock.acquire()
            self._zmq._DWX_MTX_SUBSCRIBE_MARKETDATA_(_symbol)
            print('>> DWX MESSAGE >> Subscribed to {} price feed'.format(_symbol))
              
          finally:
            # Release lock
            self._lock.release()        
            sleep(self._delay)

        # configure symbols to receive price feeds        
        try:
          # Acquire lock
          self._lock.acquire()
          self._zmq._DWX_MTX_SEND_TRACKPRICES_REQUEST_(self._symbols)
          print('>> DWX MESSAGE >> Configuring price feed for {} symbols'.format(len(self._symbols)))
            
        finally:
          # Release lock
          self._lock.release()
          sleep(self._delay)      


""" -----------------------------------------------------------------------------------------------
    -----------------------------------------------------------------------------------------------
    SCRIPT SETUP
    -----------------------------------------------------------------------------------------------
    -----------------------------------------------------------------------------------------------
"""
if __name__ == "__main__":
  import DWX_ZMQ_Strategy
  # creates object with a predefined configuration: symbol list including EURUSD and GDAXI
  print('Loading example...')
  example = prices_subscriptions(_symbols=['EURUSD','USDCAD'])  

  # Starts example execution
  print('running example...')  
  example.run()

  while not price_feed.isFinished():
        feed = price_feed.get_prices()
        if feed is None:
            break
        currency_pair = feed[0]
        prices = feed[1]
        print(f'Pair: {feed[0]}    Price: {feed[1]}')
        sleep(1)



#!/usr/bin/env python

"""
This module contains a class that is used by user to create a strategy object which is passed to fxmanager's built-in live and historical simulators.

Public Classes:
    - strategy_template: Class for injecting user defined trading strategies into fxmanager's built-in live and historical simulators.
"""

import pandas as pd
import fxmanager._metadata as md
from __main__ import __dict__

__dict__.update(md.__dict__)

class strategy_template():
    """
    Class for injecting user defined trading strategies into fxmanager's built-in live and historical simulators.

    Args:
        - strategy       : reference to a user defined function for some trading strategy. The function MUST be structured as follows in order for the simulation to work proberly:  
                            - takes ONLY ONE positional argument called 'prices'. This argument is a pandas DataFrame with range index and the following columns:
                                - ask_open: open ask price of the candle stick.
                                - bid_open: open bid price of the candle stick.
                                - ask_close: close ask price of the candle stick.
                                - bid_close: close bid price of the candle stick.
                            - takes **kwargs for any other keword arguments.
                            - returns a pandas DataFrame with ONE RAW and the following columns:
                                - order_type: can be one of 3 values (buy, sell, hold)
                                - SL : stop loss.
                                - TP : take profit.
        - take_all_prices: boolean that controls the frequency of the range index of 'prices' DataFrame.  
                           If True:
                           all the prices history since the begining of the day is passed to the strategy function, the index is a normal range index (0, 1, 2, ...). the time difference between every reading is determined by the 'sleep_time' argument in the 'fxmanager.simulation.live.run()' or 'fxmanager.simulation.historic.run()' functions.
                           If False:
                           the frequncy of the prices becomes dependant on the time frame in which we want to generate an order, for examble, if we want to open a position in 'EURUSD' in the '5min' time frame, the index will become (0, 5, 10, ..).
    Methods:
        - get_orders: used in 'fxmanager.simulation.live.run()' and 'fxmanager.simulation.historic.run()' functions.
    """

    def __init__(self, strategy, take_all_prices = False):
        self._strategy = strategy
        self._currency_pair = ''
        self._time_frame = ''
        self._take_all_prices = take_all_prices

    def kwargs_template():
        """
        returns a template dictionary for keyword args used in used defined strategies. 
        """
        ## Replace keys and values with keys and values of the arguments required for your strategy.
        kwargs = {'arg_1_key':'arg_1_value',
                  'arg_2_key':'arg_2_value',
                  'arg_3_key':'arg_3_value'
                  }

        return kwargs

    def get_portfolios_template():
        """
        returns a list of dictionaries of length = num_days where each dictionary is a portfolio used for a given day.
        """
        
        ## Repeat the dictionaries to match the number of work days in the dataset and replace the values in the lists with the desired values
        ## for each day.
        ## If you want to use the same portfolio for all the days or you are using the live simulator, just fill the first dictionary and remove
        ## other dictionaries from the portfolios list.
        portfolios = [
                    {
                        'currency_pairs' : ['pair_1', 'pair_2', 'pair_3'], # replaced by a string of currency pair symbol
                        'time_frames' : ['tf_1', 'tf2', 'tf3'], # replaced by a string of '1min', '2min' or '10min' and so on.
                        'weights' : ['w1', 'w2', 'w3'] # replaced by floats that sum up to 1 (100%).
                    },
                    {
                        'currency_pairs' : ['pair_1', 'pair_2', 'pair_3'],
                        'time_frames' : ['tf_1', 'tf2', 'tf3'],
                        'weights' : ['w1', 'w2', 'w3']
                    },
                    ]
        
        return portfolios

    def template(prices, **kwargs):
        """
        function template for the user to build trading strategies.

        Args:
            - prices: pandas DataFrame with range index and the following columns:
                - ask_open: open ask price of the candle stick.
                - bid_open: open bid price of the candle stick.
                - ask_close: close ask price of the candle stick.
                - bid_close: close bid price of the candle stick.
            - kwargs: special dictionary to hold any number of required arguments for the function.
        
        Returns:
            - orders: pandas DataFrame with ONE RAW and the following columns:
                - order_type: can be one of 3 values (buy, sell, hold)
                - SL : stop loss.
                - TP : take profit.
        """
        
        # Step 1: Extract keyword arguments from keargs dictionary
        ##########################################################
        ##########################################################

        ## Your Code Here
        # Examble: to extract a keyword called 'foo'
        # foo = kwargs['foo']

        ##########################################################
        ##########################################################
        
        # Step 2: define the logic of the strategy
        ##########################################################
        ##########################################################

        ## Your Code Here (remove oreders = [])
        orders = []

        ##########################################################
        ##########################################################
        return orders

    def get_orders(self, prices, **kwargs):
        if self._take_all_prices:
            pc = pd.DataFrame({'ask_open': prices[self._currency_pair + '_ask_open'].values,
                               'bid_open': prices[self._currency_pair + '_bid_open'].values,
                               'ask_close': prices[self._currency_pair + '_ask_close'].values,
                               'bid_close': prices[self._currency_pair + '_bid_close'].values}, index = prices.index)
        else:
            if self._time_frame[1].isalpha():
                period = int(int(self._time_frame[0]))
            else:
                period = int(int(self._time_frame[:2]))
            pc = pd.DataFrame({'ask_open': prices.loc[(prices.index % period) == 0, self._currency_pair + '_ask_open'].values,
                            'bid_open': prices.loc[(prices.index % period) == 0, self._currency_pair + '_bid_open'].values,
                            'ask_close': prices.loc[(prices.index % period) == 0, self._currency_pair + '_ask_close'].values,
                            'bid_close': prices.loc[(prices.index % period) == 0, self._currency_pair + '_bid_close'].values},
                                index = prices.index[(prices.index % period) == 0])
        orders = self._strategy(prices=pc, **kwargs)
        return orders

if __name__ == '__main__':
    pass